import functools
import logging
import re
from abc import ABC, abstractmethod
from typing import Sequence, List, Union, Callable, Any, Dict, TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

from .. import util, data_transformation
from ..columngen import ColumnGenerator
from ..data_transformation import DFTNormalisation, DFTFromFeatureGenerator, DataFrameTransformer
from ..util import flatten_arguments
from ..util.string import or_regex_group, ToStringMixin, list_string
from ..util.typing import PandasNamedTuple

if TYPE_CHECKING:
    from ..vector_model import VectorModel

log = logging.getLogger(__name__)


class DuplicateColumnNamesException(Exception):
    pass


class FeatureGenerator(ToStringMixin, ABC):
    """
    Base class for feature generators that create a new DataFrame containing feature values
    from an input DataFrame
    """
    def __init__(self,
            categorical_feature_names: Optional[Union[Sequence[str], str]] = None,
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            add_categorical_default_rules=True):
        """
        :param categorical_feature_names: either a sequence of column names or a regex that is to match all categorical feature names
            (which must not only work for the feature generated by this feature generator, i.e. it should not match feature names generated
            by other feature generators).
            It will be ensured that the respective columns in the generated data frames will have dtype 'category'.
            Furthermore, presence of meta-information can later be leveraged for further transformations, e.g. one-hot encoding.
        :param normalisation_rules: Rules to be used by DFTNormalisation (e.g. for constructing an input transformer for a model).
            These rules are only relevant if a DFTNormalisation object consuming them is instantiated and used
            within a data processing pipeline. They do not affect feature generation.
        :param normalisation_rule_template: This parameter can be supplied instead of normalisation_rules for the case where
            there shall be a single rule that applies to all columns generated by this feature generator that were not labeled as
            categorical.
        :param add_categorical_default_rules:
            If True, normalisation rules for categorical features (which are unsupported by normalisation) and their corresponding one-hot
            encoded features (with "_<index>" appended) will be added.
        """
        # NOTE: While it would be more elegant to not have all of the above constructor arguments and instead provide
        # them later using "with*" methods, this would have the significant drawback that it would enable
        # all such attributes to be provided in all subclasses, even in ones where we know settings exactly
        # and can provide them directly in the subclass constructor implementation. Thus it would enable
        # non-sensical settings which should be avoided.
        if len(normalisation_rules) > 0 and normalisation_rule_template is not None:
            raise ValueError(f"Normalisation rules should be empty when a rule template is provided")

        self._generatedColumnNames = None
        self.__categoricalFeatureNames = categorical_feature_names

        if type(categorical_feature_names) == str:
            categorical_feature_name_regex = categorical_feature_names
        else:
            if categorical_feature_names is not None and len(categorical_feature_names) > 0:
                categorical_feature_name_regex = or_regex_group(categorical_feature_names)
            else:
                categorical_feature_name_regex = None
        self._categoricalFeatureNameRegex: str = categorical_feature_name_regex
        self._categoricalFeatureRules = []

        if normalisation_rule_template is not None:
            # Note: placeholder rule's regex will be set in generate
            self._normalisationRules = [normalisation_rule_template.to_placeholder_rule()]
            self._mustUpdateNormalisationRuleBasedOnColumnNames = True
        else:
            self._normalisationRules = list(normalisation_rules)
            self._mustUpdateNormalisationRuleBasedOnColumnNames = False

        if add_categorical_default_rules:
            if categorical_feature_name_regex is not None:
                self._categoricalFeatureRules.append(data_transformation.DFTNormalisation.Rule(categorical_feature_name_regex,
                    unsupported=True))
                self._categoricalFeatureRules.append(data_transformation.DFTNormalisation.Rule(categorical_feature_name_regex + r"_\d+",
                    skip=True))  # rule for one-hot transformation

        self._name = None
        self._isFitted = False

    # for backwards compatibility with persisted Featuregens based on code prior to commit 7088cbbe
    # They lack the __isFitted attribute and we assume that each such Featuregen was fitted
    def __setstate__(self, d):
        d["_isFitted"] = d.get("_isFitted", True)
        self.__dict__ = d

    def _tostring_exclude_private(self) -> bool:
        return True

    def _tostring_additional_entries(self) -> Dict[str, Any]:
        return dict(name=self.get_name())

    def get_name(self):
        """
        :return: the name of this feature generator, which may be a default name if the name has not been set. Note that feature generators
            created by a FeatureGeneratorFactory always get the name with which the generator factory was registered.
        """
        if self._name is None:
            return f"{self.__class__.__name__}-{id(self)}"
        return self._name

    def set_name(self, name):
        self._name = name

    def get_names(self) -> list:
        """
        :return: the list of names of feature generators; will be a list with a single name for a regular feature generator
        """
        return [self.get_name()]

    def info(self):
        return {
            "name": self.get_name(),
            "categoricalFeatureNames": self.__categoricalFeatureNames,
            "generatedColumnNames": self.get_generated_column_names(),
            "isFitted": self.is_fitted(),
            "normalisationRules": self.get_normalisation_rules(),
        }

    def get_normalisation_rules(self, include_generated_categorical_rules=True) -> List[data_transformation.DFTNormalisation.Rule]:
        if include_generated_categorical_rules:
            return self._normalisationRules + self._categoricalFeatureRules
        else:
            return self._normalisationRules

    def get_categorical_feature_name_regex(self) -> Optional[str]:
        return self._categoricalFeatureNameRegex

    def is_categorical_feature(self, feature_name):
        if self._categoricalFeatureNameRegex is None:
            return False
        return re.fullmatch(self._categoricalFeatureNameRegex, feature_name) is not None

    def get_generated_column_names(self) -> Optional[List[str]]:
        """
        :return: Column names of the data frame generated by the most recent call of the feature generators 'generate' method.
            Returns None if generate was never called.
        """
        return self._generatedColumnNames

    def to_dft(self):
        return DFTFromFeatureGenerator(self)

    @abstractmethod
    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        """
        Fits the feature generator based on the given data

        :param x: the input/features data frame for the learning problem
        :param y: the corresponding output data frame for the learning problem
            (which will typically contain regression or classification target columns)
        :param ctx: a context object whose functionality may be required for feature generation;
            this is typically the model instance that this feature generator is to generate inputs for
        """
        pass

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        """
        Fits the feature generator based on the given data

        :param x: the input/features data frame for the learning problem
        :param y: the corresponding output data frame for the learning problem
            (which will typically contain regression or classification target columns)
        :param ctx: a context object whose functionality may be required for feature generation;
            this is typically the model instance that this feature generator is to generate inputs for
        """
        log.debug(f"Fitting {self}")
        self._fit(x, y=y, ctx=ctx)
        self._isFitted = True

    def is_fitted(self):
        return self._isFitted

    def generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        """
        Generates features for the data points in the given data frame

        :param df: the input data frame for which to generate features
        :param ctx: a context object whose functionality may be required for feature generation;
            this is typically the model instance that this feature generator is to generate inputs for
        :return: a data frame containing the generated features, which uses the same index as X (and Y)
        """
        if not self.is_fitted():
            raise Exception(f"Cannot generate features from a FeatureGenerator which is not fitted: "
                            f"the feature generator {self.get_name()} requires fitting")

        log.debug(f"Generating features with {self}")
        result_df = self._generate(df, ctx=ctx)

        is_column_duplicated_array = result_df.columns.duplicated()
        if any(is_column_duplicated_array):
            duplicated_columns = set(result_df.columns[is_column_duplicated_array])
            raise DuplicateColumnNamesException(f"Feature data frame contains duplicate column names: {duplicated_columns}")

        # ensure that categorical columns have dtype 'category'
        categorical_feature_names = []
        if self._categoricalFeatureNameRegex is not None:
            result_df = result_df.copy()  # result_df we got might be a view of some other DF, so before we modify it, we must copy it
            categorical_feature_names = [col for col in result_df.columns if self.is_categorical_feature(col)]
            for colName in categorical_feature_names:
                series = result_df[colName].copy()
                if series.dtype.name != 'category':
                    result_df[colName] = series.astype('category', copy=False)

        self._generatedColumnNames = result_df.columns

        # finalise normalisation rule template (if any) by making it apply to all non-categorical features
        # (a default rule applies to categorical features)
        if self._mustUpdateNormalisationRuleBasedOnColumnNames:
            non_categorical_features = list(set(self._generatedColumnNames).difference(categorical_feature_names))
            # NOTE: We here update the existing rule which was instantiated with a dummy regex because
            # some mechanisms (e.g. MultiFeatureGenerators) retrieve rule instances early on (before generate
            # is ever called) and therefore updating an existing rule is the safe route and should always
            # work, because rules should never actually be applied before generate has indeed been called
            self._normalisationRules[0].set_regex(or_regex_group(non_categorical_features))
            self._mustUpdateNormalisationRuleBasedOnColumnNames = False

        return result_df

    @abstractmethod
    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        """
        Generates features for the data points in the given data frame.

        :param df: the input data frame for which to generate features
        :param ctx: a context object whose functionality may be required for feature generation;
            this is typically the model instance that this feature generator is to generate inputs for
        :return: a data frame containing the generated features, which uses the same index as ``df``.
            The data frame's columns holding categorical columns are not required to have dtype ``category``;
            this will be ensured by the encapsulating call as long as the respective columns' names
            were appropriately provided at construction.
        """
        pass

    def fit_generate(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None) -> pd.DataFrame:
        """
        Fits the feature generator and subsequently generates features for the data points in the given data frame

        :param x: the input data frame for the learning problem and for which to generate features
        :param y: the corresponding output data frame for the learning problem
            (which will typically contain regression or classification target columns)
        :param ctx: a context object whose functionality may be required for feature generation;
            this is typically the model instance that this feature generator is to generate inputs for
        :return: a data frame containing the generated features, which uses the same index as X (and Y)
        """
        self.fit(x, y, ctx)
        return self.generate(x, ctx)

    def flattened(self,
            columns_to_flatten: List[str] = None,
            normalisation_rules=(),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            keep_other_columns=True) -> "ChainedFeatureGenerator":
        """
        Returns a new feature generator which returns flattened versions of one or more of the vector-valued columns generated
        by this feature generator.

        :param columns_to_flatten: the list of columns to flatten; if None, flatten all columns
        :param normalisation_rules: a list of normalisation rules which apply to the flattened columns
        :param normalisation_rule_template: a normalisation rule template which applies to all generated flattened columns
        :param keep_other_columns: if True, any additional columns that are not to be flattened are to be retained
            by the returned feature generator; if False, additional columns are to be discarded
        :return: a feature generator which generates the flattened columns
        """
        return flattened_feature_generator(self, columns_to_flatten=columns_to_flatten, normalisation_rules=normalisation_rules,
            keep_other_columns=keep_other_columns, normalisation_rule_template=normalisation_rule_template)

    def concat(self, *others: "FeatureGenerator") -> "MultiFeatureGenerator":
        """
        Concatenates this feature generator with one or more other feature generator in order to produce a feature generator that
        jointly generates all features

        :param others: other feature generators
        :return: a :class:`MultiFeatureGenerator`
        """
        if isinstance(self, MultiFeatureGenerator):
            fgens = list(self.featureGenerators)
        else:
            fgens = [self]
        fgens.extend(others)
        return MultiFeatureGenerator(fgens)

    def chain(self, *others: "FeatureGenerator") -> "ChainedFeatureGenerator":
        """
        Chains this feature generator with one or more other feature generators such that each feature generator
        receives as input the output of the preceding feature generator. The resulting feature generator
        produces the features of the last element in the chain.

        :param others: other feature generator
        :return: a :class:`ChainedFeatureGenerator`
        """
        if isinstance(self, ChainedFeatureGenerator):
            fgens = self.featureGenerators
        else:
            fgens = [self]
        fgens.extend(others)
        return ChainedFeatureGenerator(fgens)


class RuleBasedFeatureGenerator(FeatureGenerator, ABC):
    """
    A feature generator which does not require fitting
    """
    def fit(self, x, y=None, ctx=None):
        pass

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        pass

    def is_fitted(self):
        return True


class MultiFeatureGenerator(FeatureGenerator):
    """
    Wrapper for multiple feature generators. Calling generate here applies all given feature generators independently and
    returns the concatenation of their outputs
    """
    def __init__(self, *feature_generators: Union[FeatureGenerator, List[FeatureGenerator]]):
        self.featureGenerators = feature_generators = flatten_arguments(feature_generators)
        if len(self.featureGenerators) == 0:
            log.debug("Creating an empty MultiFeatureGenerator. It will generate a data frame without columns.")
        categorical_feature_name_regexes = [regex for regex in [fg.get_categorical_feature_name_regex()
            for fg in feature_generators] if regex is not None]
        if len(categorical_feature_name_regexes) > 0:
            categorical_feature_names = "|".join(categorical_feature_name_regexes)
        else:
            categorical_feature_names = ()
        normalisation_rules = util.concat_sequences([fg.get_normalisation_rules() for fg in feature_generators])
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            add_categorical_default_rules=False)

    def _tostring_object_info(self) -> str:
        return f"featureGenerators={list_string(self.featureGenerators)}"

    def _generate_from_multiple(self, generate_features: Callable[[FeatureGenerator], pd.DataFrame], index) -> pd.DataFrame:
        dfs = []
        for fg in self.featureGenerators:
            df = generate_features(fg)
            dfs.append(df)
        if len(dfs) == 0:
            return pd.DataFrame(index=index)
        else:
            combined_df = pd.concat(dfs, axis=1)
            if len(combined_df.columns) != len(set(combined_df.columns)):
                raise Exception(f"At least one column was generated more than once: {list(combined_df.columns)}; "
                                f"check feature generators for correctness!")
            return combined_df

    def _generate(self, input_df: pd.DataFrame, ctx=None):
        def generate_features(fg: FeatureGenerator):
            return fg.generate(input_df, ctx=ctx)
        return self._generate_from_multiple(generate_features, input_df.index)

    def fit_generate(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None) -> pd.DataFrame:
        log.debug(f"Fitting and generating features with {self}")

        def generate_features(fg: FeatureGenerator):
            return fg.fit_generate(x, y, ctx)

        return self._generate_from_multiple(generate_features, x.index)

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        for fg in self.featureGenerators:
            fg.fit(x, y)

    def is_fitted(self):
        return all([fg.is_fitted() for fg in self.featureGenerators])

    def info(self):
        info = super(MultiFeatureGenerator, self).info()
        info["featureGeneratorNames"] = self.get_names()
        return info

    def get_names(self) -> list:
        return functools.reduce(lambda x, y: x + y, [fg.get_names() for fg in self.featureGenerators], [])


class FeatureGeneratorFromNamedTuples(FeatureGenerator, ABC):
    """
    Generates feature values for one data point at a time, creating a dictionary with
    feature values from each named tuple
    """
    def __init__(self, cache: util.cache.PersistentKeyValueCache = None, categorical_feature_names: Sequence[str] = (),
                 normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
                 normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None):
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template)
        self.cache = cache

    def _generate(self, df: pd.DataFrame, ctx=None):
        dicts = []
        for idx, nt in enumerate(df.itertuples()):
            nt: PandasNamedTuple
            if idx % 100 == 0:
                log.debug(f"Generating feature via {self.__class__.__name__} for index {idx}")
            value = None
            if self.cache is not None:
                value = self.cache.get(nt.Index)
            if value is None:
                value = self._generate_feature_dict(nt)
                if self.cache is not None:
                    self.cache.set(nt.Index, value)
            dicts.append(value)
        return pd.DataFrame(dicts, index=df.index)

    @abstractmethod
    def _generate_feature_dict(self, named_tuple) -> Dict[str, Any]:
        """
        Creates a dictionary with feature values from a named tuple

        :param named_tuple: the data point for which to generate features
        :return: a dictionary mapping feature names to values
        """
        pass


class FeatureGeneratorTakeColumns(RuleBasedFeatureGenerator):
    def __init__(self, columns: Union[str, List[str]] = None, except_columns: Sequence[str] = (),
            categorical_feature_names: Optional[Union[Sequence[str], str]] = (),
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            verify_column_names=True):
        """
        :param columns: name of the column or list of names of columns to be taken. If None, all columns will be taken.
        :param except_columns: list of names of columns to not take if present in the input df
        :param categorical_feature_names: either a sequence of column names or a regex that is to match all categorical feature names
            (which must not only work for the feature generated by this feature generator, i.e. it should not match feature names generated
            by other feature generators).
            It will be ensured that the respective columns in the generated data frames will have dtype 'category'.
            Furthermore, presence of meta-information can later be leveraged for further transformations, e.g. one-hot encoding.
        :param normalisation_rules: Rules to be used by DFTNormalisation (e.g. for constructing an input transformer for a model).
            These rules are only relevant if a DFTNormalisation object consuming them is instantiated and used
            within a data processing pipeline. They do not affect feature generation.
        :param normalisation_rule_template: This parameter can be supplied instead of normalisationRules for the case where
            there shall be a single rule that applies to all columns generated by this feature generator that were not labeled as
            categorical.
        :param verify_column_names: if True and columns to take were specified, will raise an error in case said columns
            are missing during feature generation. If False, will log on info level instead
        """
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template)
        if isinstance(columns, str):
            columns = [columns]
        self.columns = columns
        self.exceptColumns = except_columns
        self.verifyColumnNames = verify_column_names

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        columns_to_take = self.columns if self.columns is not None else df.columns
        columns_to_take = [col for col in columns_to_take if col not in self.exceptColumns]

        if self.columns is not None:
            missing_cols = set(columns_to_take).difference(df.columns)
            if len(missing_cols) > 0:
                missing_cols_notification = f"Columns {missing_cols} were specified but are not present in data frame. " \
                               f"verifyColumnNames was set to {self.verifyColumnNames}; " \
                               f"available columns: {list(df.columns)}"
                if self.verifyColumnNames:
                    raise RuntimeError(missing_cols_notification)
                log.info(missing_cols_notification)
        return df[columns_to_take]

    def info(self):
        info = super().info()
        info["columns"] = self.columns
        info["exceptColumns"] = self.exceptColumns
        return info


class FeatureGeneratorFlattenColumns(RuleBasedFeatureGenerator):
    """
    Instances of this class take columns with vectors and creates a data frame with columns containing entries of
    these vectors.

    For example, if columns "vec1", "vec2" contain vectors of dimensions dim1, dim2, a data frame with dim1+dim2 new columns
    will be created. It will contain the columns "vec1_<i1>", "vec2_<i2>" with i1, i2 ranging in (0, dim1), (0, dim2).

    """
    def __init__(self, columns: Optional[Union[str, Sequence[str]]] = None, categorical_feature_names: Sequence[str] = (),
                 normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
                 normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None):
        """

        :param columns: name of the column or list of names of columns to be flattened. If None, all columns will be flattened.
        :param categorical_feature_names:
        :param normalisation_rules:
        :param normalisation_rule_template:
        """
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template)
        if isinstance(columns, str):
            columns = [columns]
        self.columns = columns

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        result_df = pd.DataFrame(index=df.index)
        columns_to_flatten = self.columns if self.columns is not None else df.columns
        for col in columns_to_flatten:
            log.debug(f"Flattening column {col}")
            # NOTE: we found the use of np.stack to produce the most runtime-efficient results.
            # Other variants, e.g. based on lists instead of numpy.arrays, perform much worse.
            values = np.stack(df[col].values)
            if len(values.shape) != 2:
                raise ValueError(f"Column {col} was expected to contain one dimensional vectors, something went wrong")
            dimension = values.shape[1]
            new_columns = [f"{col}_{i}" for i in range(dimension)]
            log.debug(f"Flattening resulted in {len(new_columns)} new columns")
            result_df[new_columns] = pd.DataFrame(values, index=df.index)
        return result_df

    def info(self):
        info = super().info()
        info["columns"] = self.columns
        return info


class FeatureGeneratorFromColumnGenerator(RuleBasedFeatureGenerator):
    """
    Implements a feature generator via a column generator
    """
    log = log.getChild(__qualname__)

    def __init__(self, column_gen: ColumnGenerator, take_input_column_if_present=False, is_categorical=False,
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None):
        """
        :param column_gen: the underlying column generator
        :param take_input_column_if_present: if True, then if a column whose name corresponds to the column to generate exists
            in the input data, simply copy it to generate the output (without using the column generator); if False, always
            apply the columnGen to generate the output
        :param is_categorical: whether the resulting column is categorical
        :param normalisation_rule_template: template for a DFTNormalisation for the resulting column.
            This should only be provided if isCategorical is False
        """
        if is_categorical and normalisation_rule_template is not None:
            raise ValueError(f"normalisationRuleTemplate should be None when the generated column is categorical")

        categorical_feature_names = (column_gen.generatedColumnName,) if is_categorical else ()
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rule_template=normalisation_rule_template)

        self.takeInputColumnIfPresent = take_input_column_if_present
        self.columnGen = column_gen

    def info(self):
        info = super().info()
        info["takeInputColumnIfPresent"] = self.takeInputColumnIfPresent
        info["generatedColName"] = self.columnGen.generatedColumnName
        return info

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        col_name = self.columnGen.generatedColumnName
        if self.takeInputColumnIfPresent and col_name in df.columns:
            self.log.debug(f"Taking column '{col_name}' from input data frame")
            series = df[col_name]
        else:
            self.log.debug(f"Generating column '{col_name}' via {self.columnGen}")
            series = self.columnGen.generate_column(df)
        return pd.DataFrame({col_name: series})


class ChainedFeatureGenerator(FeatureGenerator):
    """
    Chains feature generators such that they are executed one after another. The output of generator i>=1 is the input of
    generator i+1 in the generator sequence.
    """
    def __init__(self, *feature_generators: Union[FeatureGenerator, List[FeatureGenerator]]):
        """
        :param feature_generators: feature generators to apply in order; the properties of the last feature generator
            determine the relevant meta-data such as categorical feature names and normalisation rules
        """
        self.featureGenerators = flatten_arguments(feature_generators)
        if len(feature_generators) == 0:
            raise ValueError("Empty list of feature generators")
        last_fg: FeatureGenerator = self.featureGenerators[-1]
        super().__init__(
            categorical_feature_names=last_fg.get_categorical_feature_name_regex(), normalisation_rules=last_fg.get_normalisation_rules(),
            add_categorical_default_rules=False)

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        for featureGen in self.featureGenerators:
            df = featureGen.generate(df, ctx)
        return df

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        self.fit_generate(x, y, ctx)

    def is_fitted(self):
        return all([fg.is_fitted() for fg in self.featureGenerators])

    def info(self):
        info = super().info()
        info["chainedFeatureGeneratorNames"] = self.get_names()

    def fit_generate(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None) -> pd.DataFrame:
        log.debug(f"Fitting and generating features with {self}")
        for fg in self.featureGenerators:
            x = fg.fit_generate(x, y, ctx)
        return x


class FeatureGeneratorTargetDistribution(FeatureGenerator):
    """
    A feature generator, which, for a column T (typically the categorical target column of a classification problem
    or the continuous target column of a regression problem),

    * can ensure that T takes on limited set of values t_1, ..., t_n by allowing the user to apply
      binning using given bin boundaries
    * computes for each value c of a categorical column C the conditional empirical distribution
      P(T | C=c) in the training data during the training phase,
    * generates, for each requested column C and value c in the column, n features
      '<C>_<T>_distribution_<t_i>' = P(T=t_i | C=c) if flatten=True
      or one feature '<C>_<T>_distribution' = [P(T=t_i | C=c), ..., P(T=t_n | C=c)] if flatten=False

    Being probability values, the features generated by this feature generator are already normalised.
    """
    def __init__(self,
            columns: Union[str, Sequence[str]],
            target_column: str,
            target_column_bins: Optional[Union[Sequence[float], int, pd.IntervalIndex]],
            target_column_in_features_df=False,
            flatten=True):
        """
        :param columns: the categorical columns for which to generate distribution features
        :param target_column: the column the distributions over which will make up the features.
            If targetColumnBins is not None, this column will be discretised before computing the conditional distributions
        :param target_column_bins: if not None, specifies the binning to apply via pandas.cut
            (see https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.cut.html).
            Note that if a value should match no bin, NaN will generated. To avoid this when specifying bin boundaries in a list,
            -inf and +inf should be used as the first and last entries.
        :param target_column_in_features_df: if True, when fitting will look for targetColumn in the features data frame (X) instead of in
            target data frame (Y)
        :param flatten: whether to generate a separate scalar feature per distribution value rather than one feature
            with all of the distribution's values
        """
        self.flatten = flatten
        if isinstance(columns, str):
            columns = [columns]
        self.columns = columns
        self.targetColumn = target_column
        self.targetColumnInFeaturesDf = target_column_in_features_df
        self.targetColumnBins = target_column_bins
        if self.flatten:
            normalisation_rule_template = data_transformation.DFTNormalisation.RuleTemplate(skip=True)
        else:
            normalisation_rule_template = data_transformation.DFTNormalisation.RuleTemplate(unsupported=True)
        super().__init__(normalisation_rule_template=normalisation_rule_template)
        self._targetColumnValues = None
        # This will hold the mapping: column -> featureValue -> targetValue -> targetValueEmpiricalProbability
        self._discreteTargetDistributionsByColumn: Optional[Dict[str, Dict[Any, Dict[Any, float]]]] = None

    def info(self):
        info = super().info()
        info["columns"] = self.columns
        info["targetColumn"] = self.targetColumn
        info["targetColumnBins"] = self.targetColumnBins
        info["flatten"] = self.flatten
        return info

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        """
        This will persist the empirical target probability distributions for all unique values in the specified columns
        """
        if self.targetColumnInFeaturesDf:
            target = x[self.targetColumn]
        else:
            target = y[self.targetColumn]
        if self.targetColumnBins is not None:
            discretised_target = pd.cut(target, self.targetColumnBins)
        else:
            discretised_target = target
        self._targetColumnValues = discretised_target.unique()

        self._discreteTargetDistributionsByColumn = {}
        for column in self.columns:
            self._discreteTargetDistributionsByColumn[column] = {}
            column_target_df = pd.DataFrame()
            column_target_df[column] = x[column]
            column_target_df["target"] = discretised_target.values
            for value, valueTargetsDf in column_target_df.groupby(column):
                # The normalized value_counts contain targetValue -> targetValueEmpiricalProbability for the current value
                self._discreteTargetDistributionsByColumn[column][value] = valueTargetsDf["target"].value_counts(normalize=True).to_dict()

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        if self._discreteTargetDistributionsByColumn is None:
            raise Exception("Feature generator has not been fitted")
        result_df = pd.DataFrame(index=df.index)
        for column in self.columns:
            target_distribution_by_value = self._discreteTargetDistributionsByColumn[column]
            if self.flatten:
                for target_value in self._targetColumnValues:
                    # Important: pd.Series.apply should not be used here, as it would label the resulting column as categorical
                    result_df[f"{column}_{self.targetColumn}_distribution_{target_value}"] = \
                        [target_distribution_by_value[value].get(target_value, 0.0) for value in df[column]]
            else:
                distributions = [[target_distribution_by_value[value].get(targetValue, 0.0) for targetValue in self._targetColumnValues]
                                 for value in df[column]]
                result_df[f"{column}_{self.targetColumn}_distribution"] = pd.Series(distributions, index=df[column].index)
        return result_df


class FeatureGeneratorFromVectorModel(FeatureGenerator):
    def __init__(self,
            vector_model: "VectorModel",
            target_feature_generator: FeatureGenerator,
            categorical_feature_names: Sequence[str] = (),
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            input_feature_generator: FeatureGenerator = None,
            use_target_feature_generator_for_training=False):
        """
        Provides a feature via predictions of a given model

        :param vector_model: model used for generate features from predictions
        :param target_feature_generator: generator for target to be predicted
        :param categorical_feature_names:
        :param normalisation_rules:
        :param normalisation_rule_template:
        :param input_feature_generator: optional feature generator to be applied to input of vectorModel's fit and predict
        :param use_target_feature_generator_for_training: if False, this generator will always apply the model
            to generate features.
            If True, this generator will use targetFeatureGenerator to generate features, bypassing the
            model. This is useful for the case where the model which is
            to receive the generated features shall be trained on the original targets rather than the predictions
            thereof.
        """
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template)

        self.useTargetFeatureGeneratorForTraining = use_target_feature_generator_for_training
        self.targetFeatureGenerator = target_feature_generator
        self.inputFeatureGenerator = input_feature_generator
        self.useTargetFeatureGeneratorForTraining = use_target_feature_generator_for_training
        self.vectorModel = vector_model

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        target_df = self.targetFeatureGenerator.fit_generate(x, y)
        if self.inputFeatureGenerator:
            x = self.inputFeatureGenerator.fit_generate(x, y)
        self.vectorModel.fit(x, target_df)

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        if self.inputFeatureGenerator:
            df = self.inputFeatureGenerator.generate(df)
        if self.useTargetFeatureGeneratorForTraining and not ctx.is_fitted():
            log.debug(f"Using targetFeatureGenerator {self.targetFeatureGenerator.__class__.__name__} to generate target features")
            return self.targetFeatureGenerator.generate(df)
        else:
            log.debug(f"Generating target features via {self.vectorModel.__class__.__name__}")
            return self.vectorModel.predict(df)

    def info(self):
        info = super().info()
        info["wrappedModel"] = str(self.vectorModel)
        return info


class FeatureGeneratorMapColumn(RuleBasedFeatureGenerator, ABC):
    """
    Creates a single feature from a single input column by applying a function to each element of the input column
    """
    def __init__(self,
            input_col_name: str,
            feature_col_name: str,
            categorical_feature_names: Optional[Union[Sequence[str], str]] = None,
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            add_categorical_default_rules=True):
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template, add_categorical_default_rules=add_categorical_default_rules)
        self._inputColName = input_col_name
        self._featureColName = feature_col_name

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        if self._inputColName not in df.columns:
            raise ValueError(f"Column '{self._inputColName}' required by feature generator not found in list of columns: "
                             f"{list(df.columns)}")
        input_series = df[self._inputColName]
        values = input_series.apply(self._create_value)
        return pd.DataFrame({self._featureColName: values}, index=df.index)

    @abstractmethod
    def _create_value(self, value):
        """
        Maps a value from the input column to a feature value

        :param value: a value from the input column
        :return: the feature value
        """
        pass


class FeatureGeneratorMapColumnDict(RuleBasedFeatureGenerator, ABC):
    """
    Creates an arbitrary number of features from a single input column by applying a function to each element of the input column
    """
    def __init__(self, input_col_name: str, categorical_feature_names: Optional[Union[Sequence[str], str]] = None,
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None, add_categorical_default_rules=True):
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template, add_categorical_default_rules=add_categorical_default_rules)
        self._inputColName = input_col_name

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        if self._inputColName not in df.columns:
            raise ValueError(f"Column '{self._inputColName}' required by feature generator not found in list of columns: "
                             f"{list(df.columns)}")
        input_series = df[self._inputColName]
        values = [self._create_features_dict(v) for v in input_series]
        return pd.DataFrame(values, index=df.index)

    @abstractmethod
    def _create_features_dict(self, value) -> Dict[str, Any]:
        """
        Maps a value from the input column to a dictionary containing one or more features.

        :param value: a value from the input column
        :return: a dictionary mapping feature names to values
        """
        pass


class FeatureGeneratorNAMarker(RuleBasedFeatureGenerator):
    """
    Creates features indicating whether another feature is N/A (not available).
    It can be practical to use this feature generator in conjunction with DFTFillNA for models that cannot handle missing values.
    """
    def __init__(self, columns: List[str], value_a=0, value_na=1):
        """
        Note: When changing the default values used, use only values that are considered to be normalised when using this
        feature generation in a context where DFTNormalisation is used (no normalisation is applied to features generated
        by this feature generator).

        :param columns: the columns for which to generate
        :param value_a: the feature value if the input feature is available
        :param value_na: the feature value if the input feature is not available
        """
        super().__init__(normalisation_rule_template=DFTNormalisation.RuleTemplate(skip=True))
        self.columns = columns
        self.valueA = value_a
        self.valueNA = value_na

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        new_cols = {}
        value_map = {True: self.valueNA, False: self.valueA}
        for col in self.columns:
            new_cols[f"{col}_na"] = [value_map[isNA] for isNA in df[col].isna()]
        return pd.DataFrame(new_cols, index=df.index)


def flattened_feature_generator(fgen: FeatureGenerator, columns_to_flatten: List[str] = None, keep_other_columns=True,
        normalisation_rules: Sequence[DFTNormalisation.Rule] = (),
        normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None):
    """
    Return a flattening version of the input feature generator.

    :param fgen: the feature generator which generates columns that are to be flattened
    :param columns_to_flatten: list of names of output columns to be flattened; if None, flatten all columns
    :param keep_other_columns: whether any additional columns that are not to be flattened are to be retained
        by the returned feature generator
    :param normalisation_rules: additional normalisation rules for the flattened output columns
    :param normalisation_rule_template: This parameter can be supplied instead of normalisationRules for the case where
        there shall be a single rule that applies to all flattened output columns
    :return: FeatureGenerator instance that will generate flattened versions of the specified columns and leave
        all other output columns as is.
    
    Example:
        >>> from sensai.featuregen import FeatureGeneratorTakeColumns, flattened_feature_generator
        >>> import pandas as pd
        >>>
        >>> df = pd.DataFrame({"foo": [[1, 2], [3, 4]], "bar": ["a", "b"]})
        >>> fgen = flattened_feature_generator(FeatureGeneratorTakeColumns(), columns_to_flatten=["foo"])
        >>> fgen.generate(df)
           foo_0  foo_1 bar
        0      1      2   a
        1      3      4   b
    """
    flattening_generator = FeatureGeneratorFlattenColumns(columns=columns_to_flatten, normalisation_rules=normalisation_rules,
        normalisation_rule_template=normalisation_rule_template)
    if columns_to_flatten is None or not keep_other_columns:
        return ChainedFeatureGenerator(fgen, flattening_generator)
    else:
        return ChainedFeatureGenerator(fgen,
            MultiFeatureGenerator(flattening_generator, FeatureGeneratorTakeColumns(except_columns=columns_to_flatten)))


class FeatureGeneratorFromDFT(FeatureGenerator):
    def __init__(self, dft: DataFrameTransformer, categorical_feature_names: Optional[Union[Sequence[str], str]] = None,
            normalisation_rules: Sequence[data_transformation.DFTNormalisation.Rule] = (),
            normalisation_rule_template: data_transformation.DFTNormalisation.RuleTemplate = None,
            add_categorical_default_rules=True):
        super().__init__(categorical_feature_names=categorical_feature_names, normalisation_rules=normalisation_rules,
            normalisation_rule_template=normalisation_rule_template, add_categorical_default_rules=add_categorical_default_rules)
        self.dft = dft

    def _fit(self, x: pd.DataFrame, y: pd.DataFrame = None, ctx=None):
        self.dft.fit(x)

    def _generate(self, df: pd.DataFrame, ctx=None) -> pd.DataFrame:
        return self.dft.apply(df)



import logging
from abc import ABC, abstractmethod
from typing import Union, Set, Callable, Iterable, Optional

import numpy as np
import pandas as pd
from scipy.spatial import distance_matrix

from ..util.cache import PickleLoadSaveMixin

log = logging.getLogger(__name__)


# TODO at some point in the future: generalize to other input and deal with algorithms that allow prediction of labels
class EuclideanClusterer(PickleLoadSaveMixin, ABC):
    """
    Base class for all clustering algorithms. Supports noise clusters and relabelling of identified clusters as noise
    based on their size.

    :param noise_label: label that is associated with the noise cluster or None
    :param min_cluster_size: if not None, clusters below this size will be labeled as noise
    :param max_cluster_size: if not None, clusters above this size will be labeled as noise
    """
    def __init__(self, noise_label=-1, min_cluster_size: int = None, max_cluster_size: int = None):
        self._datapoints: Optional[np.ndarray] = None
        self._labels: Optional[np.ndarray] = None
        self._clusterIdentifiers: Optional[Set[int]] = None
        self._nonNoiseClusterIdentifiers: Optional[Set[int]] = None

        if min_cluster_size is not None or max_cluster_size is not None:
            if noise_label is None:
                raise ValueError("the noise label has to be not None for non-trivial bounds on cluster sizes")
        self.noiseLabel = noise_label
        self.maxClusterSize = max_cluster_size if max_cluster_size is not None else np.inf
        self.minClusterSize = min_cluster_size if min_cluster_size is not None else -np.inf

        self._clusterDict = {}
        self._numClusters: Optional[int] = None

    class Cluster:
        def __init__(self, datapoints: np.ndarray, identifier: Union[int, str]):
            self.datapoints = datapoints
            self.identifier = identifier
            self._radius: Optional[float] = None
            self._centroid: Optional[np.ndarray] = None

        def __len__(self):
            return len(self.datapoints)

        def __str__(self):
            return f"{self.__class__.__name__}_{self.identifier}"

        def _compute_radius(self):
            return np.max(distance_matrix([self.centroid()], self.datapoints))

        def _compute_centroid(self):
            return np.mean(self.datapoints, axis=0)

        def centroid(self):
            if self._centroid is None:
                self._centroid = self._compute_centroid()
            return self._centroid

        def radius(self):
            if self._radius is None:
                self._radius = self._compute_radius()
            return self._radius

        def summary_dict(self):
            """
            :return: dictionary containing coarse information about the cluster (e.g. num_members and centroid)
            """
            return {
                "identifier": self.identifier,
                "centroid": self.centroid(),
                "numMembers": len(self),
                "radius": self.radius()
            }

    @classmethod
    def __str__(cls):
        return cls.__name__

    def clusters(self, condition: Callable[[Cluster], bool] = None) -> Iterable[Cluster]:
        """
        :param condition: if provided, only clusters fulfilling the condition will be included
        :return: generator of clusters
        """
        percentage_to_log = 0
        for i, clusterId in enumerate(self._nonNoiseClusterIdentifiers):
            # logging process through the loop
            percentage_generated = int(100 * i / self.num_clusters)
            if percentage_generated == percentage_to_log:
                log.info(f"Processed {percentage_to_log}% of clusters")
                percentage_to_log += 5

            cluster = self.get_cluster(clusterId)
            if condition is None or condition(cluster):
                yield cluster

    def noise_cluster(self):
        if self.noiseLabel is None:
            raise NotImplementedError(f"The algorithm {self} does not provide a noise cluster")
        return self.get_cluster(self.noiseLabel)

    def summary_df(self, condition: Callable[[Cluster], bool] = None):
        """
        :param condition: if provided, only clusters fulfilling the condition will be included
        :return: pandas DataFrame containing coarse information about the clusters
        """
        summary_dicts = [cluster.summary_dict() for cluster in self.clusters(condition=condition)]
        return pd.DataFrame(summary_dicts).set_index("identifier", drop=True)

    def fit(self, data: np.ndarray) -> None:
        log.info(f"Fitting {self} to {len(data)} coordinate datapoints.")
        labels = self._compute_labels(data)
        if len(labels) != len(data):
            raise Exception(f"Bad Implementation: number of labels does not match number of datapoints")
        # Relabel clusters that do not fulfill size bounds as noise
        if self.minClusterSize != -np.inf or self.maxClusterSize != np.inf:
            for clusterId, clusterSize in zip(*np.unique(labels, return_counts=True)):
                if not self.minClusterSize <= clusterSize <= self.maxClusterSize:
                    labels[labels == clusterId] = self.noiseLabel

        self._datapoints = data
        self._clusterIdentifiers = set(labels)
        self._labels = labels
        if self.noiseLabel is not None:
            self._nonNoiseClusterIdentifiers = self._clusterIdentifiers.difference({self.noiseLabel})
        log.info(f"{self} found {self.num_clusters} clusters")

    @property
    def is_fitted(self):
        return self._datapoints is not None

    @property
    def datapoints(self) -> np.ndarray:
        assert self.is_fitted
        return self._datapoints

    @property
    def labels(self) -> np.ndarray:
        assert self.is_fitted
        return self._labels

    @property
    def cluster_identifiers(self) -> Set[int]:
        assert self.is_fitted
        return self._clusterIdentifiers

    # unfortunately, there seems to be no way to annotate the return type correctly
    # https://github.com/python/mypy/issues/3993
    def get_cluster(self, cluster_id: int) -> Cluster:
        if cluster_id not in self.labels:
            raise KeyError(f"no cluster for id {cluster_id}")
        result = self._clusterDict.get(cluster_id)
        if result is None:
            result = self.Cluster(self.datapoints[self.labels == cluster_id], identifier=cluster_id)
            self._clusterDict[cluster_id] = result
        return result

    @property
    def num_clusters(self) -> int:
        return len(self._nonNoiseClusterIdentifiers)

    @abstractmethod
    def _compute_labels(self, x: np.ndarray) -> np.ndarray:
        """
        Fit the clustering model and return an array of integer cluster labels

        :param x: the datapoints
        :return: list of the same length as the input datapoints; it represents the mapping coordinate -> cluster_id
        """
        pass



import collections
import itertools
import math
from abc import abstractmethod, ABC
from typing import List, Tuple, Iterator, Optional

import numpy as np
import sklearn.cluster

from .geo_coords import GeoCoord
from .local_coords import LocalCoordinateSystem
from ..clustering import GreedyAgglomerativeClustering


class GeoCoordClusterer(ABC):
    @abstractmethod
    def fit_geo_coords(self, geo_coords: List[GeoCoord]):
        """
        :param geo_coords: the coordinates to be clustered
        """
        pass

    @abstractmethod
    def clusters_indices(self) -> Tuple[List[List[int]], List[int]]:
        """
        :return: a tuple (clusters, outliers), where clusters is a list of point indices, one list for each
            cluster containing the indices of points within the cluster, and outliers is the list of indices of points not within
            clusters
        """
        pass


class GreedyAgglomerativeGeoCoordClusterer(GeoCoordClusterer):
    def __init__(self,
            max_min_distance_for_merge_m: float,
            max_distance_m: float,
            min_cluster_size: int,
            lcs: LocalCoordinateSystem = None):
        """
        :param max_min_distance_for_merge_m: the maximum distance, in metres, for the minimum distance between two existing clusters for a merge
            to be admissible
        :param max_distance_m: the maximum distance, in metres, between any two points for the points to be allowed to be in the same cluster
        :param min_cluster_size: the minimum number of points any valid cluster must ultimately contain; the points in any smaller clusters
            shall be considered as outliers
        :param lcs: the local coordinate system to use for clustering; if None, compute based on mean coordinates passed when fitting
        """
        self.lcs = lcs
        self.min_cluster_size = min_cluster_size
        self.max_min_distance_for_merge = max_min_distance_for_merge_m
        self.max_distance_m = max_distance_m
        self.squared_max_min_distance_for_merge = max_min_distance_for_merge_m * max_min_distance_for_merge_m
        self.squared_max_distance = max_distance_m * max_distance_m
        self.local_points = None
        self.m_min_distance: Optional["GreedyAgglomerativeGeoCoordClusterer.Matrix"] = None
        self.m_max_squared_distance: Optional["GreedyAgglomerativeGeoCoordClusterer.Matrix"] = None
        self.clusters = None

    class Matrix:
        UNSET_VALUE = np.inf

        def __init__(self, dim: int):
            self.m = np.empty((dim, dim))
            self.m.fill(np.inf)

        def set(self, c1: int, c2: int, value: float):
            self.m[c1][c2] = value
            self.m[c2][c1] = value

        def get(self, c1: int, c2: int) -> float:
            return self.m[c1][c2]

    class LocalPoint:
        def __init__(self, xy: np.ndarray, idx: int):
            self.idx = idx
            self.xy = xy

    class Cluster(GreedyAgglomerativeClustering.Cluster):
        def __init__(self, point: "GreedyAgglomerativeGeoCoordClusterer.LocalPoint", idx: int,
                clusterer: 'GreedyAgglomerativeGeoCoordClusterer'):
            self.idx = idx
            self.clusterer = clusterer
            self.points = [point]

        def merge_cost(self, other: "GreedyAgglomerativeGeoCoordClusterer.Cluster"):
            cartesian_product = itertools.product(self.points, other.points)
            min_squared_distance = math.inf
            max_squared_distance = 0
            for p1, p2 in cartesian_product:
                diff = p1.xy - p2.xy
                squared_distance = np.dot(diff, diff)
                if squared_distance > self.clusterer.squared_max_distance:
                    max_squared_distance = math.inf
                    break
                else:
                    min_squared_distance = min(squared_distance, min_squared_distance)

            # fill cache: the max value takes precedence; if it is inf (no merge admissible), then the min value is also set to inf;
            # the min value valid only if the max value is finite
            self.clusterer.m_max_squared_distance.set(self.idx, other.idx, max_squared_distance)
            if np.isinf(max_squared_distance):
                self.clusterer.m_min_distance.set(self.idx, other.idx, np.inf)
            else:
                self.clusterer.m_min_distance.set(self.idx, other.idx, min_squared_distance)

            if np.isinf(max_squared_distance):
                return math.inf
            if min_squared_distance <= self.clusterer.squared_max_min_distance_for_merge:
                return min_squared_distance
            return math.inf

        def merge(self, other):
            self.points += other.points

    def fit_geo_coords(self, geo_coords: List[GeoCoord]) -> None:
        self.m_min_distance = self.Matrix(len(geo_coords))
        self.m_max_squared_distance = self.Matrix(len(geo_coords))
        if self.lcs is None:
            mean_coord = GeoCoord.mean_coord(geo_coords)
            self.lcs = LocalCoordinateSystem(mean_coord.lat, mean_coord.lon)
        self.local_points = [self.LocalPoint(np.array(self.lcs.get_local_coords(p.lat, p.lon)), idx) for idx, p in enumerate(geo_coords)]
        clusters = [self.Cluster(lp, i, self) for i, lp in enumerate(self.local_points)]
        gac = GreedyAgglomerativeClustering(clusters,
            merge_candidate_determination_strategy=self.MergeCandidateDeterminationStrategy(self.max_distance_m, self))
        clusters = gac.apply_clustering()
        self.clusters = clusters

    def clusters_indices(self) -> Tuple[List[List[int]], List[int]]:
        outliers = []
        clusters = []
        for c in self.clusters:
            indices = [p.idx for p in c.points]
            if len(c.points) < self.min_cluster_size:
                outliers.extend(indices)
            else:
                clusters.append(indices)
        return clusters, outliers

    class MergeCandidateDeterminationStrategy(GreedyAgglomerativeClustering.MergeCandidateDeterminationStrategy):
        def __init__(self, search_radius_m: float, parent: "GreedyAgglomerativeGeoCoordClusterer"):
            super().__init__()
            self.parent = parent
            self.searchRadiusM = search_radius_m

        def set_clusterer(self, clusterer: GreedyAgglomerativeClustering):
            super().set_clusterer(clusterer)
            points = []
            for wc in self.clusterer.wrapped_clusters:
                c: GreedyAgglomerativeGeoCoordClusterer.Cluster = wc.cluster
                for p in c.points:
                    points.append(p.xy)
            assert len(points) == len(self.clusterer.wrapped_clusters)
            points = np.stack(points)
            self.kdtree = sklearn.neighbors.KDTree(points)

        def iter_candidate_indices(self, wc: "GreedyAgglomerativeClustering.WrappedCluster", initial: bool,
                merged_cluster_indices: Tuple[int, int] = None) -> Iterator[int]:
            c: GreedyAgglomerativeGeoCoordClusterer.Cluster = wc.cluster
            if initial:
                local_point = c.points[0]  # pick any point from wc, since we use maximum cluster extension as search radius
                indices = self.kdtree.query_radius(np.reshape(local_point.xy, (1, 2)), self.searchRadiusM)[0]
                candidate_set = set()
                for idx in indices:
                    wc = self.clusterer.wrapped_clusters[idx]
                    candidate_set.add(wc.get_cluster_association().idx)
                yield from sorted(candidate_set)
            else:
                # The new distance values (max/min) between wc and any cluster index otherIdx can be computed from the cached distance
                # values of the two clusters from which wc was created through a merge:
                # The max distance is the maximum of the squared distances of the original clusters (and if either is inf, then
                # a merge is definitely inadmissible, because one of the original clusters was already too far away).
                # The min distance is the minimum of the squred distances of the original clusters.
                c1, c2 = merged_cluster_indices
                max1 = self.parent.m_max_squared_distance.m[c1]
                max2 = self.parent.m_max_squared_distance.m[c2]
                max_combined = np.maximum(max1, max2)
                for otherIdx, maxSqDistance in enumerate(max_combined):
                    min_sq_distance = np.inf
                    if maxSqDistance <= self.parent.squared_max_distance:
                        wc_other = self.clusterer.wrapped_clusters[otherIdx]
                        if wc_other.is_merged():
                            continue
                        min1 = self.parent.m_min_distance.get(c1, otherIdx)
                        min2 = self.parent.m_min_distance.get(c2, otherIdx)
                        min_sq_distance = min(min1, min2)
                        if min_sq_distance <= self.parent.squared_max_min_distance_for_merge:
                            yield GreedyAgglomerativeClustering.ClusterMerge(wc, wc_other, min_sq_distance)
                    # update cache
                    self.parent.m_max_squared_distance.set(wc.idx, otherIdx, maxSqDistance)
                    self.parent.m_min_distance.set(wc.idx, otherIdx, min_sq_distance)


class SkLearnGeoCoordClusterer(GeoCoordClusterer):
    def __init__(self, clusterer, lcs: LocalCoordinateSystem = None):
        """
        :param clusterer: a clusterer from sklearn.cluster
        :param lcs: the local coordinate system to use for Euclidian conversion; if None, determine from data (using mean coordinate as
            centre)
        """
        self.lcs = lcs
        self.clusterer = clusterer
        self.local_points = None

    def fit_geo_coords(self, geo_coords: List[GeoCoord]):
        if self.lcs is None:
            mean_coord = GeoCoord.mean_coord(geo_coords)
            self.lcs = LocalCoordinateSystem(mean_coord.lat, mean_coord.lon)
        self.local_points = [self.lcs.get_local_coords(p.lat, p.lon) for p in geo_coords]
        self.clusterer.fit(self.local_points)

    def _clusters(self, mode):
        clusters = collections.defaultdict(list)
        outliers = []
        for idxPoint, idxCluster in enumerate(self.clusterer.labels_):
            if mode == "localPoints":
                item = self.local_points[idxPoint]
            elif mode == "indices":
                item = idxPoint
            else:
                raise ValueError()
            if idxCluster >= 0:
                clusters[idxCluster].append(item)
            else:
                outliers.append(item)
        return list(clusters.values()), outliers

    def clusters_local_points(self) -> Tuple[List[List[Tuple[float, float]]], List[Tuple[float, float]]]:
        """
        :return: a tuple (clusters, outliers), where clusters is a dictionary mapping from cluster index to
            the list of local points within the cluster and outliers is a list of local points not within
            clusters
        """
        return self._clusters("localPoints")

    def clusters_indices(self) -> Tuple[List[List[int]], List[int]]:
        return self._clusters("indices")


class DBSCANGeoCoordClusterer(SkLearnGeoCoordClusterer):
    def __init__(self, eps, min_samples, lcs: LocalCoordinateSystem = None, **kwargs):
        """
        :param eps: the maximum distance between two samples for one to be considered as in the neighbourhood of the other
        :param min_samples: the minimum number of samples that must be within a neighbourhood for a cluster to be formed
        :param lcs: the local coordinate system for conversion to a Euclidian space
        :param kwargs: additional arguments to pass to DBSCAN (see https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html)
        """
        super().__init__(sklearn.cluster.DBSCAN(eps=eps, min_samples=min_samples, **kwargs), lcs)



from typing import Optional

import xgboost

from .sklearn.sklearn_base import AbstractSkLearnMultipleOneDimVectorRegressionModel, AbstractSkLearnVectorClassificationModel, \
    FeatureImportanceProviderSkLearnRegressionMultipleOneDim, FeatureImportanceProviderSkLearnClassification


def is_xgboost_version_at_least(major: int, minor: Optional[int] = None, patch: Optional[int] = None):
    components = xgboost.__version__.split(".")
    for i, version in enumerate((major, minor, patch)):
        if version is not None:
            installed_version = int(components[i])
            if installed_version > version:
                return True
            if installed_version < version:
                return False
    return True


class XGBGradientBoostedVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    """
    XGBoost's regression model using gradient boosted trees
    """
    def __init__(self, random_state=42, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRegressor
        """
        super().__init__(xgboost.XGBRegressor, random_state=random_state, **model_args)


class XGBRandomForestVectorRegressionModel(AbstractSkLearnMultipleOneDimVectorRegressionModel,
        FeatureImportanceProviderSkLearnRegressionMultipleOneDim):
    """
    XGBoost's random forest regression model
    """
    def __init__(self, random_state=42, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRFRegressor
        """
        super().__init__(xgboost.XGBRFRegressor, random_state=random_state, **model_args)


class XGBGradientBoostedVectorClassificationModel(AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification):
    """
    XGBoost's classification model using gradient boosted trees
    """
    def __init__(self, random_state=42, use_balanced_class_weights=False, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBClassifier
        """
        use_label_encoding = is_xgboost_version_at_least(1, 6)
        super().__init__(xgboost.XGBClassifier, random_state=random_state, use_balanced_class_weights=use_balanced_class_weights,
            use_label_encoding=use_label_encoding, **model_args)


class XGBRandomForestVectorClassificationModel(AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification):
    """
    XGBoost's random forest classification model
    """
    def __init__(self, random_state=42, use_balanced_class_weights=False, **model_args):
        """
        :param model_args: See https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRFClassifier
        """
        use_label_encoding = is_xgboost_version_at_least(1, 6)
        super().__init__(xgboost.XGBRFClassifier, random_state=random_state, use_balanced_class_weights=use_balanced_class_weights,
            use_label_encoding=use_label_encoding, **model_args)



import enum
import functools
import logging
import math
import time
from abc import ABC, abstractmethod
from collections import OrderedDict
from enum import Enum
from typing import List, Union, Sequence, Callable, TYPE_CHECKING, Tuple, Optional, Dict, Any

import matplotlib.figure
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.optim as optim
from matplotlib import pyplot as plt
from torch import cuda as torchcuda

from .torch_data import TensorScaler, DataUtil, TorchDataSet, TorchDataSetProviderFromDataUtil, TorchDataSetProvider, \
    TensorScalerIdentity, TensorTransformer
from .torch_enums import ClassificationOutputMode
from ..util.string import ToStringMixin

if TYPE_CHECKING:
    from .torch_base import TorchModel

log = logging.getLogger(__name__)


class Optimiser(enum.Enum):
    SGD = ("sgd", optim.SGD)
    ASGD = ("asgd", optim.ASGD)
    ADAGRAD = ("adagrad", optim.Adagrad)
    ADADELTA = ("adadelta", optim.Adadelta)
    ADAM = ("adam", optim.Adam)
    ADAMW = ("adamw", optim.AdamW)
    ADAMAX = ("adamax", optim.Adamax)
    RMSPROP = ("rmsprop", optim.RMSprop)
    RPROP = ("rprop", optim.Rprop)
    LBFGS = ("lbfgs", optim.LBFGS)

    @classmethod
    def from_name(cls, name: str) -> "Optimiser":
        lname = name.lower()
        for o in cls:
            if o.value[0] == lname:
                return o
        raise ValueError(f"Unknown optimiser name '{name}'; known names: {[o.value[0] for o in cls]}")

    @classmethod
    def from_name_or_instance(cls, name_or_instance: Union[str, "Optimiser"]) -> "Optimiser":
        if type(name_or_instance) == str:
            return cls.from_name(name_or_instance)
        else:
            return name_or_instance


class _Optimiser(object):
    """
    Wrapper for classes inherited from torch.optim.Optimizer
    """
    def __init__(self, params, method: Union[str, Optimiser], lr, max_grad_norm, use_shrinkage=True, **optimiser_args):
        """
        :param params: an iterable of torch.Tensor s or dict s. Specifies what Tensors should be optimized.
        :param method: the optimiser to use
        :param lr: learnig rate
        :param max_grad_norm: gradient norm value beyond which to apply gradient shrinkage
        :param optimiser_args: keyword arguments to be used in actual torch optimiser
        """
        self.method = Optimiser.from_name_or_instance(method)
        self.params = list(params)  # careful: params may be a generator
        self.last_ppl = None
        self.lr = lr
        self.max_grad_norm = max_grad_norm
        self.start_decay = False
        self.optimiserArgs = optimiser_args
        self.use_shrinkage = use_shrinkage

        # instantiate optimiser
        optimiser_args = dict(self.optimiserArgs)
        optimiser_args.update({'lr': self.lr})
        if self.method == Optimiser.LBFGS:
            self.use_shrinkage = False
            self.optimizer = optim.LBFGS(self.params, **optimiser_args)
        else:
            cons = self.method.value[1]
            self.optimizer = cons(self.params, **optimiser_args)

    def step(self, loss_backward: Callable):
        """
        :param loss_backward: callable, performs backward step and returns loss
        :return: loss value
        """
        if self.use_shrinkage:
            def closure_with_shrinkage():
                loss_value = loss_backward()
                torch.nn.utils.clip_grad_norm_(self.params, self.max_grad_norm)
                return loss_value

            closure = closure_with_shrinkage
        else:
            closure = loss_backward

        loss = self.optimizer.step(closure)
        return loss


class NNLossEvaluator(ABC):
    """
    Base class defining the interface for training and validation loss evaluation.
    """
    class Evaluation(ABC):
        @abstractmethod
        def start_epoch(self) -> None:
            """
            Starts a new epoch, resetting any aggregated values required to ultimately return the
            epoch's overall training loss (via getEpochTrainLoss) and validation metrics (via getValidationMetrics)
            """
            pass

        @abstractmethod
        def compute_train_batch_loss(self, model_output, ground_truth, x, y) -> torch.Tensor:
            """
            Computes the loss for the given model outputs and ground truth values for a batch
            and aggregates the computed loss values such that :meth:``getEpochTrainLoss`` can return an appropriate
            result for the entire epoch.
            The original batch tensors X and Y are provided as meta-information only.

            :param model_output: the model output
            :param ground_truth: the ground truth values
            :param x: the original batch input tensor
            :param y: the original batch output (ground truth) tensor
            :return: the loss (scalar tensor)
            """
            pass

        @abstractmethod
        def get_epoch_train_loss(self) -> float:
            """
            :return: the epoch's overall training loss (as obtained by collecting data from individual training
                batch data passed to computeTrainBatchLoss)
            """
            pass

        @abstractmethod
        def process_validation_batch(self, model_output, ground_truth, x, y) -> None:
            """
            Processes the given model outputs and ground truth values in order to compute sufficient statistics for
            velidation metrics, which at the end of the epoch, shall be retrievable via method getValidationMetrics

            :param model_output: the model output
            :param ground_truth: the ground truth values
            :param x: the original batch input tensor
            :param y: the original batch output (ground truth) tensor
            :return: the loss (scalar tensor)
            """
            pass

        @abstractmethod
        def get_validation_metrics(self) -> Dict[str, float]:
            pass

    @abstractmethod
    def start_evaluation(self, cuda: bool) -> Evaluation:
        """
        Begins the evaluation of a model, returning a (stateful) object which is to perform the necessary computations.

        :param cuda: whether CUDA is being applied (all tensors/models on the GPU)
        :return: the evaluation object
        """
        pass

    @abstractmethod
    def get_validation_metric_name(self) -> str:
        """
        :return: the name of the validation metric which is to be used to determine the best model (key for the ordered
            dictionary returned by method Evaluation.getValidationMetrics)
        """
        pass


class NNLossEvaluatorFixedDim(NNLossEvaluator, ABC):
    """
    Base class defining the interface for training and validation loss evaluation, which uses fixed-dimension
    outputs and aggregates individual training batch losses that are summed losses per batch
    (averaging appropriately internally).
    """
    class Evaluation(NNLossEvaluator.Evaluation):
        def __init__(self, criterion, validation_loss_evaluator: "NNLossEvaluatorFixedDim.ValidationLossEvaluator",
                output_dim_weights: torch.Tensor = None):
            self.output_dim_weights = output_dim_weights
            self.output_dim_weight_sum = torch.sum(output_dim_weights) if output_dim_weights is not None else None
            self.validation_loss_evaluator = validation_loss_evaluator
            self.criterion = criterion
            self.total_loss = None
            self.num_samples = None
            self.num_outputs_per_data_point: Optional[int] = None
            self.validation_ground_truth_shape = None

        def start_epoch(self):
            self.total_loss = 0
            self.num_samples = 0
            self.validation_ground_truth_shape = None

        def compute_train_batch_loss(self, model_output, ground_truth, x, y) -> torch.Tensor:
            # size of modelOutput and groundTruth: (batchSize, outputDim=numOutputsPerDataPoint)
            if self.num_outputs_per_data_point is None:
                output_shape = y.shape[1:]
                self.num_outputs_per_data_point = functools.reduce(lambda x, y: x * y, output_shape, 1)
                assert self.output_dim_weights is None or len(self.output_dim_weights) == self.num_outputs_per_data_point
            num_data_points_in_batch = y.shape[0]
            if self.output_dim_weights is None:
                # treat all dimensions as equal, applying criterion to entire tensors
                loss = self.criterion(model_output, ground_truth)
                self.num_samples += num_data_points_in_batch * self.num_outputs_per_data_point
                self.total_loss += loss.item()
                return loss
            else:
                # compute loss per dimension and return weighted loss
                loss_per_dim = torch.zeros(self.num_outputs_per_data_point, device=model_output.device, dtype=torch.float)
                for o in range(self.num_outputs_per_data_point):
                    loss_per_dim[o] = self.criterion(model_output[:, o], ground_truth[:, o])
                weighted_loss = (loss_per_dim * self.output_dim_weights).sum() / self.output_dim_weight_sum
                self.num_samples += num_data_points_in_batch
                self.total_loss += weighted_loss.item()
                return weighted_loss

        def get_epoch_train_loss(self) -> float:
            return self.total_loss / self.num_samples

        def process_validation_batch(self, model_output, ground_truth, x, y):
            if self.validation_ground_truth_shape is None:
                self.validation_ground_truth_shape = y.shape[1:]  # the shape of the output of a single model application
                self.validation_loss_evaluator.start_validation_collection(self.validation_ground_truth_shape)
            self.validation_loss_evaluator.process_validation_result_batch(model_output, ground_truth)

        def get_validation_metrics(self) -> Dict[str, float]:
            return self.validation_loss_evaluator.end_validation_collection()

    def start_evaluation(self, cuda: bool) -> Evaluation:
        criterion = self.get_training_criterion()
        output_dim_weights_array = self.get_output_dim_weights()
        output_dim_weights_tensor = torch.from_numpy(output_dim_weights_array).float() if output_dim_weights_array is not None else None
        if cuda:
            criterion = criterion.cuda()
            if output_dim_weights_tensor is not None:
                output_dim_weights_tensor = output_dim_weights_tensor.cuda()
        return self.Evaluation(criterion, self.create_validation_loss_evaluator(cuda), output_dim_weights=output_dim_weights_tensor)

    @abstractmethod
    def get_training_criterion(self) -> nn.Module:
        """
        Gets the optimisation criterion (loss function) for training.
        Standard implementations are available in torch.nn (torch.nn.MSELoss, torch.nn.CrossEntropyLoss, etc.).
        """
        pass

    @abstractmethod
    def get_output_dim_weights(self) -> Optional[np.ndarray]:
        pass

    @abstractmethod
    def create_validation_loss_evaluator(self, cuda: bool) -> "ValidationLossEvaluator":
        """
        :param cuda: whether to use CUDA-based tensors
        :return: the evaluator instance which is to be used to evaluate the model on validation data
        """
        pass

    def get_validation_metric_name(self) -> str:
        """
        Gets the name of the metric (key of dictionary as returned by the validation loss evaluator's
        endValidationCollection method), which is defining for the quality of the model and thus determines which
        epoch's model is considered the best.

        :return: the name of the metric
        """
        pass

    class ValidationLossEvaluator(ABC):
        @abstractmethod
        def start_validation_collection(self, ground_truth_shape):
            """
            Initiates validation data collection for a new epoch, appropriately resetting this object's internal state.

            :param ground_truth_shape: the tensor shape of a single ground truth data point (not including the batch
                entry dimension)
            """
            pass

        @abstractmethod
        def process_validation_result_batch(self, output, ground_truth):
            """
            Collects, for validation, the given output and ground truth data (tensors holding data on one batch,
            where the first dimension is the batch entry)

            :param output: the model's output
            :param ground_truth: the corresponding ground truth
            """
            pass

        @abstractmethod
        def end_validation_collection(self) -> OrderedDict:
            """
            Computes validation metrics based on the data previously processed.

            :return: an ordered dictionary with validation metrics
            """
            pass


class NNLossEvaluatorRegression(NNLossEvaluatorFixedDim, ToStringMixin):
    """A loss evaluator for (multi-variate) regression."""

    class LossFunction(Enum):
        L1LOSS = "L1Loss"
        L2LOSS = "L2Loss"
        MSELOSS = "MSELoss"
        SMOOTHL1LOSS = "SmoothL1Loss"

    def __init__(self, loss_fn: LossFunction = LossFunction.L2LOSS, validation_tensor_transformer: Optional[TensorTransformer] = None,
            output_dim_weights: Sequence[float] = None, apply_output_dim_weights_in_validation=True,
            validation_metric_name: Optional[str] = None):
        """
        :param loss_fn: the loss function to use
        :param validation_tensor_transformer: a transformer which is to be applied to validation tensors (both model outputs and ground
            truth) prior to computing the validation metrics
        :param output_dim_weights: vector of weights to apply to then mean loss per output dimension, i.e. for the case where for each data
            point, the model produces n output dimensions, the mean loss for the i-th dimension is to be computed separately and be scaled
            with the weight, and the overall loss returned is the weighted average. The weights need not sum to 1 (normalisation is
            applied).
        :param apply_output_dim_weights_in_validation: whether output dimension weights are also to be applied to to the metrics computed
            for validation. Note that this may not be possible if a validationTensorTransformer which changes the output dimensions is
            used.
        :param validation_metric_name: the metric to use for model selection during validation; if None, use default depending on lossFn
        """
        self.validation_tensor_transformer = validation_tensor_transformer
        self.output_dim_weights = np.array(output_dim_weights) if output_dim_weights is not None else None
        self.apply_output_dim_weights_in_validation = apply_output_dim_weights_in_validation
        self.validation_metric_name = validation_metric_name
        if loss_fn is None:
            loss_fn = self.LossFunction.L2LOSS
        try:
            self.loss_fn = self.LossFunction(loss_fn)
        except ValueError:
            raise Exception(f"The loss function '{loss_fn}' is not supported. "
                            f"Available options are: {[e.value for e in self.LossFunction]}")

    def create_validation_loss_evaluator(self, cuda):
        return self.ValidationLossEvaluator(cuda, self.validation_tensor_transformer, self.output_dim_weights,
            self.apply_output_dim_weights_in_validation)

    def get_training_criterion(self):
        if self.loss_fn is self.LossFunction.L1LOSS:
            criterion = nn.L1Loss(reduction='sum')
        elif self.loss_fn is self.LossFunction.L2LOSS or self.loss_fn == self.LossFunction.MSELOSS:
            criterion = nn.MSELoss(reduction='sum')
        elif self.loss_fn is self.LossFunction.SMOOTHL1LOSS:
            criterion = nn.SmoothL1Loss(reduction='sum')
        else:
            raise AssertionError(f"Loss function {self.loss_fn} defined but instantiation not implemented.")
        return criterion

    def get_output_dim_weights(self) -> Optional[np.ndarray]:
        return self.output_dim_weights

    class ValidationLossEvaluator(NNLossEvaluatorFixedDim.ValidationLossEvaluator):
        def __init__(self, cuda: bool, validation_tensor_transformer: Optional[TensorTransformer], output_dim_weights: np.ndarray,
                apply_output_dim_weights: bool):
            self.validationTensorTransformer = validation_tensor_transformer
            self.outputDimWeights = output_dim_weights
            self.applyOutputDimWeights = apply_output_dim_weights
            self.total_loss_l1 = None
            self.total_loss_l2 = None
            self.output_dims = None
            self.allTrueOutputs = None
            self.evaluate_l1 = nn.L1Loss(reduction='sum')
            self.evaluate_l2 = nn.MSELoss(reduction='sum')
            if cuda:
                self.evaluate_l1 = self.evaluate_l1.cuda()
                self.evaluate_l2 = self.evaluate_l2.cuda()
            self.begin_new_validation_collection: Optional[bool] = None

        def start_validation_collection(self, ground_truth_shape):
            if len(ground_truth_shape) != 1:
                raise ValueError("Outputs that are not vectors are currently unsupported")
            self.begin_new_validation_collection = True

        def process_validation_result_batch(self, output, ground_truth):
            # apply tensor transformer (if any)
            if self.validationTensorTransformer is not None:
                output = self.validationTensorTransformer.transform(output)
                ground_truth = self.validationTensorTransformer.transform(ground_truth)

            # check if new collection
            if self.begin_new_validation_collection:
                self.output_dims = ground_truth.shape[-1]
                self.total_loss_l1 = np.zeros(self.output_dims)
                self.total_loss_l2 = np.zeros(self.output_dims)
                self.allTrueOutputs = None
                self.begin_new_validation_collection = False

            assert len(output.shape) == 2 and len(ground_truth.shape) == 2

            # obtain series of outputs per output dimension: (batch_size, output_size) -> (output_size, batch_size)
            predicted_output = output.permute(1, 0)
            true_output = ground_truth.permute(1, 0)

            if self.allTrueOutputs is None:
                self.allTrueOutputs = true_output
            else:
                self.allTrueOutputs = torch.cat((self.allTrueOutputs, true_output), dim=1)

            for i in range(self.output_dims):
                self.total_loss_l1[i] += self.evaluate_l1(predicted_output[i], true_output[i]).item()
                self.total_loss_l2[i] += self.evaluate_l2(predicted_output[i], true_output[i]).item()

        def end_validation_collection(self):
            output_dims = self.output_dims
            rae = np.zeros(output_dims)
            rrse = np.zeros(output_dims)
            mae = np.zeros(output_dims)
            mse = np.zeros(output_dims)

            for i in range(output_dims):
                mean = torch.mean(self.allTrueOutputs[i])
                ref_model_errors = self.allTrueOutputs[i] - mean
                ref_model_sum_abs_errors = torch.sum(torch.abs(ref_model_errors)).item()
                ref_model_sum_squared_errors = torch.sum(ref_model_errors * ref_model_errors).item()
                num_samples = ref_model_errors.size(0)

                mae[i] = self.total_loss_l1[i] / num_samples
                mse[i] = self.total_loss_l2[i] / num_samples
                rae[i] = self.total_loss_l1[i] / ref_model_sum_abs_errors if ref_model_sum_abs_errors != 0 else np.inf
                rrse[i] = np.sqrt(mse[i]) / np.sqrt(
                    ref_model_sum_squared_errors / num_samples) if ref_model_sum_squared_errors != 0 else np.inf

            def mean(x):
                if self.applyOutputDimWeights:
                    return np.average(x, weights=self.outputDimWeights)
                else:
                    return np.mean(x)

            metrics = OrderedDict([("RRSE", mean(rrse)), ("RAE", mean(rae)), ("MSE", mean(mse)), ("MAE", mean(mae))])
            return metrics

    def get_validation_metric_name(self):
        if self.validation_metric_name is not None:
            return self.validation_metric_name
        else:
            if self.loss_fn is self.LossFunction.L1LOSS or self.loss_fn is self.LossFunction.SMOOTHL1LOSS:
                return "MAE"
            elif self.loss_fn is self.LossFunction.L2LOSS or self.loss_fn is self.LossFunction.MSELOSS:
                return "MSE"
            else:
                raise AssertionError(f"No validation metric defined as selection criterion for loss function {self.loss_fn}")


class NNLossEvaluatorClassification(NNLossEvaluatorFixedDim):
    """A loss evaluator for classification"""

    class LossFunction(Enum):
        CROSSENTROPY = "CrossEntropy"
        NLL = "NegativeLogLikelihood"

        def create_criterion(self) -> Callable:
            if self is self.CROSSENTROPY:
                return nn.CrossEntropyLoss(reduction='sum')
            elif self is self.NLL:
                return nn.NLLLoss(reduction="sum")

        def get_validation_metric_key(self) -> str:
            if self is self.CROSSENTROPY:
                return "CE"
            elif self is self.NLL:
                return "NLL"

        @classmethod
        def default_for_output_mode(cls, output_mode: ClassificationOutputMode):
            if output_mode == ClassificationOutputMode.PROBABILITIES:
                raise ValueError(f"No loss function available for {output_mode}; Either apply log at the end and use "
                                 f"{ClassificationOutputMode.LOG_PROBABILITIES} or use a different final activation (e.g. log_softmax) "
                                 f"to avoid this type of output.")
            elif output_mode == ClassificationOutputMode.LOG_PROBABILITIES:
                return cls.NLL
            elif output_mode == ClassificationOutputMode.UNNORMALISED_LOG_PROBABILITIES:
                return cls.CROSSENTROPY
            else:
                raise ValueError(f"No default specified for {output_mode}")

    def __init__(self, loss_fn: LossFunction):
        self.lossFn: "NNLossEvaluatorClassification.LossFunction" = self.LossFunction(loss_fn)

    def __str__(self):
        return f"{self.__class__.__name__}[{self.lossFn}]"

    def create_validation_loss_evaluator(self, cuda):
        return self.ValidationLossEvaluator(cuda, self.lossFn)

    def get_training_criterion(self):
        return self.lossFn.create_criterion()

    def get_output_dim_weights(self) -> Optional[np.ndarray]:
        return None

    class ValidationLossEvaluator(NNLossEvaluatorFixedDim.ValidationLossEvaluator):
        def __init__(self, cuda: bool, loss_fn: "NNLossEvaluatorClassification.LossFunction"):
            self.loss_fn = loss_fn
            self.total_loss = None
            self.num_validation_samples = None
            self.criterion = self.loss_fn.create_criterion()
            if cuda:
                self.criterion = self.criterion.cuda()

        def start_validation_collection(self, ground_truth_shape):
            self.total_loss = 0
            self.num_validation_samples = 0

        def process_validation_result_batch(self, output, ground_truth):
            self.total_loss += self.criterion(output, ground_truth).item()
            self.num_validation_samples += output.shape[0]

        def end_validation_collection(self):
            mean_loss = self.total_loss / self.num_validation_samples
            if isinstance(self.criterion, nn.CrossEntropyLoss):
                metrics = OrderedDict([("CE", mean_loss), ("GeoMeanProbTrueClass", math.exp(-mean_loss))])
            elif isinstance(self.criterion, nn.NLLLoss):
                metrics = {"NLL": mean_loss}
            else:
                raise ValueError()
            return metrics

    def get_validation_metric_name(self):
        return self.lossFn.get_validation_metric_key()


class NNOptimiserParams(ToStringMixin):
    REMOVED_PARAMS = {"cuda"}
    RENAMED_PARAMS = {
        "optimiserClip": "optimiser_clip",
        "lossEvaluator": "loss_evaluator",
        "optimiserLR": "optimiser_lr",
        "earlyStoppingEpochs": "early_stopping_epochs",
        "batchSize": "batch_size",
        "trainFraction": "train_fraction",
        "scaledOutputs": "scaled_outputs",
        "useShrinkage": "use_shrinkage",
        "shrinkageClip": "shrinkage_clip",
    }

    def __init__(self,
            loss_evaluator: NNLossEvaluator = None,
            gpu: Optional[int] = None,
            optimiser: Union[str, Optimiser] = "adam",
            optimiser_lr=0.001,
            early_stopping_epochs=None,
            batch_size=None,
            epochs=1000,
            train_fraction=0.75,
            scaled_outputs=False,
            use_shrinkage=True,
            shrinkage_clip=10.,
            shuffle=True,
            optimiser_args: Optional[Dict[str, Any]] = None):
        """
        :param loss_evaluator: the loss evaluator to use
        :param gpu: the index of the GPU to be used (if CUDA is enabled for the model to be trained); if None, default to first GPU
        :param optimiser: the optimiser to use
        :param optimiser_lr: the optimiser's learning rate
        :param early_stopping_epochs: the number of epochs without validation score improvement after which to abort training and
            use the best epoch's model (early stopping); if None, never abort training before all epochs are completed
        :param batch_size: the batch size to use; for algorithms L-BFGS (optimiser='lbfgs'), which do not use batches, leave this at None.
            If the algorithm uses batches and None is specified, batch size 64 will be used by default.
        :param train_fraction: the fraction of the data used for training (with the remainder being used for validation).
            If no validation is to be performed, pass 1.0.
        :param scaled_outputs: whether to scale all outputs, resulting in computations of the loss function based on scaled values rather
            than normalised values.
            Enabling scaling may not be appropriate in cases where there are multiple outputs on different scales/with completely different
            units.
        :param use_shrinkage: whether to apply shrinkage to gradients whose norm exceeds ``shrinkageClip``, scaling the gradient down to
            ``shrinkageClip``
        :param shrinkage_clip: the maximum gradient norm beyond which to apply shrinkage (if ``useShrinkage`` is True)
        :param shuffle: whether to shuffle the training data
        :param optimiser_args: keyword arguments to be passed on to the actual torch optimiser
        """
        if Optimiser.from_name_or_instance(optimiser) == Optimiser.LBFGS:
            large_batch_size = 1e12
            if batch_size is not None:
                log.warning(f"LBFGS does not make use of batches, therefore using large batch size {large_batch_size} "
                            f"to achieve use of a single batch")
            batch_size = large_batch_size
        else:
            if batch_size is None:
                log.debug("No batch size was specified, using batch size 64 by default")
                batch_size = 64

        self.epochs = epochs
        self.batch_size = batch_size
        self.optimiser_lr = optimiser_lr
        self.shrinkage_clip = shrinkage_clip
        self.optimiser = optimiser
        self.gpu = gpu
        self.train_fraction = train_fraction
        self.scaled_outputs = scaled_outputs
        self.loss_evaluator = loss_evaluator
        self.optimiser_args = optimiser_args if optimiser_args is not None else {}
        self.use_shrinkage = use_shrinkage
        self.early_stopping_epochs = early_stopping_epochs
        self.shuffle = shuffle

    @classmethod
    def _updated_params(cls, params: dict) -> dict:
        return {cls.RENAMED_PARAMS.get(k, k): v for k, v in params.items() if k not in cls.REMOVED_PARAMS}

    def __setstate__(self, state):
        if "shuffle" not in state:
            state["shuffle"] = True
        self.__dict__ = self._updated_params(state)

    @classmethod
    def from_dict_or_instance(cls, nn_optimiser_params: Union[dict, "NNOptimiserParams"]) -> "NNOptimiserParams":
        if isinstance(nn_optimiser_params, NNOptimiserParams):
            return nn_optimiser_params
        else:
            return cls.from_dict(nn_optimiser_params)

    @classmethod
    def from_dict(cls, params: dict) -> "NNOptimiserParams":
        return NNOptimiserParams(**cls._updated_params(params))

    # TODO remove deprecated dict interface
    @classmethod
    def from_either_dict_or_instance(cls, nn_optimiser_dict_params: dict, nn_optimiser_params: Optional["NNOptimiserParams"]):
        have_instance = nn_optimiser_params is not None
        have_dict = len(nn_optimiser_dict_params)
        if have_instance and have_dict:
            raise ValueError("Received both a non-empty dictionary and an instance")
        if have_instance:
            return nn_optimiser_params
        else:
            return NNOptimiserParams.from_dict(nn_optimiser_dict_params)


class NNOptimiser:
    log = log.getChild(__qualname__)

    def __init__(self, params: NNOptimiserParams):
        """
        :param params: parameters
        """
        if params.loss_evaluator is None:
            raise ValueError("Must provide a loss evaluator")

        self.params = params
        self.cuda = None
        self.best_epoch = None

    def __str__(self):
        return f"{self.__class__.__name__}[params={self.params}]"

    def fit(self,
            model: "TorchModel",
            data: Union[DataUtil, List[DataUtil], TorchDataSetProvider, List[TorchDataSetProvider],
                        TorchDataSet, List[TorchDataSet], Tuple[TorchDataSet, TorchDataSet], List[Tuple[TorchDataSet, TorchDataSet]]],
            create_torch_module=True) -> "TrainingInfo":
        """
        Fits the parameters of the given model to the given data, which can be a list of or single instance of one of the following:

            * a `DataUtil` or `TorchDataSetProvider` (from which a training set and validation set will be obtained according to
              the `trainFraction` parameter of this object)
            * a `TorchDataSet` which shall be used as the training set (for the case where no validation set shall be used)
            * a tuple with two `TorchDataSet` instances, where the first shall be used as the training set and the second as
              the validation set

        :param model: the model to be fitted
        :param data: the data to use (see variants above)
        :param create_torch_module: whether to newly create the torch module that is to be trained from the model's factory.
            If False, (re-)train the existing module.
        """
        self.cuda = model.cuda
        self.log.info(f"Preparing parameter learning of {model} via {self} with cuda={self.cuda}")

        use_validation = self.params.train_fraction != 1.0

        def to_data_set_provider(d) -> TorchDataSetProvider:
            if isinstance(d, TorchDataSetProvider):
                return d
            elif isinstance(d, DataUtil):
                return TorchDataSetProviderFromDataUtil(d, self.cuda)
            else:
                raise ValueError(f"Cannot create a TorchDataSetProvider from {d}")

        training_log_entries = []

        def training_log(s):
            self.log.info(s)
            training_log_entries.append(s)

        self._init_cuda()

        # Set the random seed manually for reproducibility.
        seed = 42
        torch.manual_seed(seed)
        if self.cuda:
            torchcuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

        # obtain data, splitting it into training and validation set(s)
        validation_sets = []
        training_sets = []
        output_scalers = []
        if type(data) != list:
            data = [data]
        self.log.info("Obtaining input/output training instances")
        for idx_data_item, data_item in enumerate(data):
            if isinstance(data_item, TorchDataSet):
                if use_validation:
                    raise ValueError("Passing a TorchDataSet instance is not admissible when validation is enabled (trainFraction != 1.0). "
                                     "Pass a TorchDataSetProvider or another representation that supports validation instead.")
                training_set = data_item
                validation_set = None
                output_scaler = TensorScalerIdentity()
            elif type(data_item) == tuple:
                training_set, validation_set = data_item
                output_scaler = TensorScalerIdentity()
            else:
                data_set_provider = to_data_set_provider(data_item)
                training_set, validation_set = data_set_provider.provide_split(self.params.train_fraction)
                output_scaler = data_set_provider.get_output_tensor_scaler()
            training_sets.append(training_set)
            if validation_set is not None:
                validation_sets.append(validation_set)
            output_scalers.append(output_scaler)
            training_log(f"Data set {idx_data_item+1}/{len(data)}: #train={training_set.size()}, "
                         f"#validation={validation_set.size() if validation_set is not None else 'None'}")
        training_log("Number of validation sets: %d" % len(validation_sets))

        torch_model = model.create_torch_module() if create_torch_module else model.get_torch_module()
        if self.cuda:
            torch_model.cuda()
        model.set_torch_module(torch_model)

        n_params = sum([p.nelement() for p in torch_model.parameters()])
        self.log.info(f"Learning parameters of {model}")
        training_log('Number of parameters: %d' % n_params)
        training_log(f"Starting training process via {self}")

        loss_evaluator = self.params.loss_evaluator

        total_epochs = None
        best_val = 1e9
        best_epoch = 0
        optim = _Optimiser(torch_model.parameters(), method=self.params.optimiser, lr=self.params.optimiser_lr,
            max_grad_norm=self.params.shrinkage_clip, use_shrinkage=self.params.use_shrinkage, **self.params.optimiser_args)

        best_model_bytes = model.get_module_bytes()
        loss_evaluation = loss_evaluator.start_evaluation(self.cuda)
        validation_metric_name = loss_evaluator.get_validation_metric_name()
        training_loss_values = []
        validation_metric_values = []
        try:
            self.log.info(f'Begin training with cuda={self.cuda}')
            self.log.info('Press Ctrl+C to end training early')
            for epoch in range(1, self.params.epochs + 1):
                loss_evaluation.start_epoch()
                epoch_start_time = time.time()

                # perform training step, processing all the training data once
                train_loss = self._train(training_sets, torch_model, optim, loss_evaluation, self.params.batch_size, output_scalers)
                training_loss_values.append(train_loss)

                # perform validation, computing the mean metrics across all validation sets (if more than one),
                # and check for new best result according to validation results
                is_new_best = False
                if use_validation:
                    metrics_sum = None
                    metrics_keys = None
                    for i, (validation_set, output_scaler) in enumerate(zip(validation_sets, output_scalers)):
                        metrics = self._evaluate(validation_set, torch_model, loss_evaluation, output_scaler)
                        metrics_array = np.array(list(metrics.values()))
                        if i == 0:
                            metrics_sum = metrics_array
                            metrics_keys = metrics.keys()
                        else:
                            metrics_sum += metrics_array
                    metrics_sum /= len(validation_sets)  # mean results
                    metrics = dict(zip(metrics_keys, metrics_sum))
                    current_val = metrics[loss_evaluator.get_validation_metric_name()]
                    validation_metric_values.append(current_val)
                    is_new_best = current_val < best_val
                    if is_new_best:
                        best_val = current_val
                        best_epoch = epoch
                        best_str = "best {:s} {:5.6f} from this epoch".format(validation_metric_name, best_val)
                    else:
                        best_str = "best {:s} {:5.6f} from epoch {:d}".format(validation_metric_name, best_val, best_epoch)
                    val_str = f' | validation {", ".join(["%s %5.4f" % e for e in metrics.items()])} | {best_str}'
                else:
                    val_str = ""
                training_log(
                    'Epoch {:3d}/{} completed in {:5.2f}s | train loss {:5.4f}{:s}'.format(
                        epoch, self.params.epochs, (time.time() - epoch_start_time), train_loss, val_str))
                total_epochs = epoch
                if use_validation:
                    if is_new_best:
                        best_model_bytes = model.get_module_bytes()

                    # check for early stopping
                    num_epochs_without_improvement = epoch - best_epoch
                    if self.params.early_stopping_epochs is not None and \
                            num_epochs_without_improvement >= self.params.early_stopping_epochs:
                        training_log(f"Stopping early: {num_epochs_without_improvement} epochs without validation metric improvement")
                        break

            training_log("Training complete")
        except KeyboardInterrupt:
            training_log('Exiting from training early because of keyboard interrupt')

        # reload best model according to validation results
        if use_validation:
            training_log(f'Best model is from epoch {best_epoch} with {validation_metric_name} {best_val} on validation set')
            self.best_epoch = best_epoch
            model.set_module_bytes(best_model_bytes)

        return TrainingInfo(best_epoch=best_epoch if use_validation else None, log=training_log_entries, total_epochs=total_epochs,
            training_loss_sequence=training_loss_values, validation_metric_sequence=validation_metric_values)

    def _apply_model(self, model, input: Union[torch.Tensor, Sequence[torch.Tensor]], ground_truth, output_scaler: TensorScaler):
        if isinstance(input, torch.Tensor):
            output = model(input)
        else:
            output = model(*input)
        if self.params.scaled_outputs:
            output, ground_truth = self._scaled_values(output, ground_truth, output_scaler)
        return output, ground_truth

    @classmethod
    def _scaled_values(cls, model_output, ground_truth, output_scaler):
        scaled_output = output_scaler.denormalise(model_output)
        scaled_truth = output_scaler.denormalise(ground_truth)
        return scaled_output, scaled_truth

    def _train(self, data_sets: Sequence[TorchDataSet], model: nn.Module, optim: _Optimiser,
            loss_evaluation: NNLossEvaluator.Evaluation, batch_size: int, output_scalers: Sequence[TensorScaler]):
        """Performs one training epoch"""
        model.train()
        for data_set, output_scaler in zip(data_sets, output_scalers):
            for X, Y in data_set.iter_batches(batch_size, shuffle=self.params.shuffle):
                def closure():
                    model.zero_grad()
                    output, ground_truth = self._apply_model(model, X, Y, output_scaler)
                    loss = loss_evaluation.compute_train_batch_loss(output, ground_truth, X, Y)
                    loss.backward()
                    return loss

                optim.step(closure)
        return loss_evaluation.get_epoch_train_loss()

    def _evaluate(self, data_set: TorchDataSet, model: nn.Module, loss_evaluation: NNLossEvaluator.Evaluation,
            output_scaler: TensorScaler):
        """Evaluates the model on the given data set (a validation set)"""
        model.eval()
        for X, Y in data_set.iter_batches(self.params.batch_size, shuffle=False):
            with torch.no_grad():
                output, ground_truth = self._apply_model(model, X, Y, output_scaler)
            loss_evaluation.process_validation_batch(output, ground_truth, X, Y)
        return loss_evaluation.get_validation_metrics()

    def _init_cuda(self):
        """Initialises CUDA (for learning) by setting the appropriate device if necessary"""
        if self.cuda:
            device_count = torchcuda.device_count()
            if device_count == 0:
                raise Exception("CUDA is enabled but no device found")
            if self.params.gpu is None:
                if device_count > 1:
                    log.warning("More than one GPU detected but no GPU index was specified, using GPU 0 by default.")
                gpu_index = 0
            else:
                gpu_index = self.params.gpu
            torchcuda.set_device(gpu_index)
        elif torchcuda.is_available():
            self.log.info("NOTE: You have a CUDA device; consider running with cuda=True")


class TrainingInfo:
    def __init__(self, best_epoch: int = None, log: List[str] = None, training_loss_sequence: Sequence[float] = None,
            validation_metric_sequence: Sequence[float] = None, total_epochs=None):
        self.validation_metric_sequence = validation_metric_sequence
        self.training_loss_sequence = training_loss_sequence
        self.log = log
        self.best_epoch = best_epoch
        self.total_epochs = total_epochs

    def __setstate__(self, state):
        if "totalEpochs" not in state:
            state["totalEpochs"] = None
        self.__dict__ = state

    def get_training_loss_series(self) -> pd.Series:
        return self._create_series_with_one_based_index(self.training_loss_sequence, name="training loss")

    def get_validation_metric_series(self) -> pd.Series:
        return self._create_series_with_one_based_index(self.validation_metric_sequence, name="validation metric")

    def _create_series_with_one_based_index(self, sequence: Sequence, name: str):
        series = pd.Series(sequence, name=name)
        series.index += 1
        return series

    def plot_all(self) -> matplotlib.figure.Figure:
        """
        Plots both the sequence of training loss values and the sequence of validation metric values
        """
        ts = self.get_training_loss_series()
        vs = self.get_validation_metric_series()

        fig, primary_ax = plt.subplots(1, 1)
        secondary_ax = primary_ax.twinx()

        training_line = primary_ax.plot(ts, color='blue')
        validation_line = secondary_ax.plot(vs, color='orange')
        best_epoc_line = primary_ax.axvline(self.best_epoch, color='black', linestyle='dashed')

        primary_ax.set_xlabel("epoch")
        primary_ax.set_ylabel(ts.name)
        secondary_ax.set_ylabel(vs.name)

        primary_ax.legend(training_line + validation_line + [best_epoc_line], [ts.name, vs.name, "best epoch"])
        plt.tight_layout()

        return fig



import logging
from typing import Union, Optional

import numpy as np
import sklearn.ensemble
import sklearn.naive_bayes
import sklearn.neural_network
import sklearn.tree
from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

from .sklearn_base import AbstractSkLearnVectorClassificationModel, FeatureImportanceProviderSkLearnClassification

log = logging.getLogger(__name__)


class SkLearnDecisionTreeVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, min_samples_leaf=1, random_state=42, **model_args):
        super().__init__(DecisionTreeClassifier,
            min_samples_leaf=min_samples_leaf, random_state=random_state, **model_args)


class SkLearnRandomForestVectorClassificationModel(AbstractSkLearnVectorClassificationModel,
        FeatureImportanceProviderSkLearnClassification):
    def __init__(self, n_estimators=100, min_samples_leaf=1, random_state=42, use_balanced_class_weights=False, **model_args):
        super().__init__(RandomForestClassifier,
            random_state=random_state, min_samples_leaf=min_samples_leaf, n_estimators=n_estimators,
            use_balanced_class_weights=use_balanced_class_weights,
            **model_args)


class SkLearnMLPVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, hidden_layer_sizes=(100,), activation: str = "relu",
            solver: str = "adam", batch_size: Union[int, str] = "auto", random_state: Optional[int] = 42,
            max_iter: int = 200, early_stopping: bool = False, n_iter_no_change: int = 10, **model_args):
        """
        :param hidden_layer_sizes: the sequence of hidden layer sizes
        :param activation: {"identity", "logistic", "tanh", "relu"} the activation function to use for hidden layers (the one used for the
            output layer is always 'identity')
        :param solver: {"adam", "lbfgs", "sgd"} the name of the solver to apply
        :param batch_size: the batch size or "auto" for min(200, data set size)
        :param random_state: the random seed for reproducability; use None if it shall not be specifically defined
        :param max_iter: the number of iterations (gradient steps for L-BFGS, epochs for other solvers)
        :param early_stopping: whether to use early stopping (stop training after n_iter_no_change epochs without improvement)
        :param n_iter_no_change: the number of iterations after which to stop early (if early_stopping is enabled)
        :param model_args: additional arguments to pass on to MLPClassifier, see
            https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html
        """
        super().__init__(sklearn.neural_network.MLPClassifier, hidden_layer_sizes=hidden_layer_sizes, activation=activation,
            random_state=random_state, solver=solver, batch_size=batch_size, max_iter=max_iter, early_stopping=early_stopping,
            n_iter_no_change=n_iter_no_change, **model_args)


class SkLearnMultinomialNBVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, **model_args):
        super().__init__(sklearn.naive_bayes.MultinomialNB, **model_args)


class SkLearnSVCVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, random_state=42, **model_args):
        super().__init__(sklearn.svm.SVC, random_state=random_state, **model_args)


class SkLearnLogisticRegressionVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, random_state=42, **model_args):
        super().__init__(sklearn.linear_model.LogisticRegression, random_state=random_state, **model_args)


class SkLearnKNeighborsVectorClassificationModel(AbstractSkLearnVectorClassificationModel):
    def __init__(self, **model_args):
        super().__init__(sklearn.neighbors.KNeighborsClassifier, **model_args)

    def _predict_sklearn(self, input_values):
        # Apply a transformation to fix a bug in sklearn 1.3.0 (and perhaps earlier versions):
        # https://github.com/scikit-learn/scikit-learn/issues/26768
        inputs = np.ascontiguousarray(input_values)

        return super()._predict_sklearn(inputs)



# -*- coding: utf-8 -*-
import logging
import math
import queue
from abc import ABC, abstractmethod
from typing import List, Sequence, Iterator, Callable, Optional, Union, Tuple

log = logging.getLogger(__name__)


class GreedyAgglomerativeClustering(object):
    """
    An implementation of greedy agglomerative clustering which avoids unnecessary 
    recomputations of merge costs through the management of a priority queue of 
    potential merges.

    Greedy agglomerative clustering works as follows. Starting with an initial
    set of clusters (where each cluster typically contains a single data point),
    the method successively merges the two clusters where the merge cost is lowest (greedy),
    until no further merges are admissible.
    The merge operation is a mutating operation, i.e. the initial clusters are modified.

    To apply the method, the Cluster class must be subclassed, so as to define
    what the cost of a merge in your application shall be and how two clusters can be merged.
    For example, if data points are points in a Cartesian coordinate system, then the merge cost
    can be defined as the minimum or maximum distance among all pairs of points in the two clusters,
    admissibility being determined by a threshold that must not be exceeded;
    the merge operation can simply concatenate lists of data points.
    """
    log = log.getChild(__qualname__)

    class Cluster(ABC):
        """
        Base class for clusters that can be merged via GreedyAgglomerativeClustering
        """
        @abstractmethod
        def merge_cost(self, other) -> float:
            """
            Computes the cost of merging the given cluster with this cluster

            :return: the (non-negative) merge cost or math.inf if a merge is inadmissible"""
            pass

        @abstractmethod
        def merge(self, other):
            """
            Merges the given cluster into this cluster"

            :param other: the cluster that is to be merged into this cluster
            """
            pass

    def __init__(self, clusters: Sequence[Cluster],
            merge_candidate_determination_strategy: "GreedyAgglomerativeClustering.MergeCandidateDeterminationStrategy" = None):
        """
        :param clusters: the initial clusters, which are to be agglomerated into larger clusters
        """
        self.prioritised_merges = queue.PriorityQueue()
        self.wrapped_clusters = []
        for idx, c in enumerate(clusters):
            self.wrapped_clusters.append(GreedyAgglomerativeClustering.WrappedCluster(c, idx, self))

        # initialise merge candidate determination strategy
        if merge_candidate_determination_strategy is None:
            merge_candidate_determination_strategy = self.MergeCandidateDeterminationStrategyDefault()
        merge_candidate_determination_strategy.set_clusterer(self)
        self.mergeCandidateDeterminationStrategy = merge_candidate_determination_strategy
        
    def apply_clustering(self) -> List[Cluster]:
        """
        Applies greedy agglomerative clustering to the clusters given at construction, merging
        clusters until no further merges are admissible

        :return: the list of agglomerated clusters (subset of the original clusters, which may have had other
            clusters merged into them)
        """
        # compute all possible merges, adding them to the priority queue
        self.log.debug("Computing initial merges")
        for idx, wc in enumerate(self.wrapped_clusters):
            self.log.debug("Computing potential merges for cluster index %d" % idx)
            wc.compute_merges(True)
        
        # greedily apply the least-cost merges
        steps = 0
        while not self.prioritised_merges.empty():
            self.log.debug("Clustering step %d" % (steps+1))
            have_merge = False
            while not have_merge and not self.prioritised_merges.empty():
                merge = self.prioritised_merges.get()
                if not merge.evaporated:
                    have_merge = True
            if have_merge:
                merge.apply()
            steps += 1
        
        result = filter(lambda wc: not wc.is_merged(), self.wrapped_clusters)
        result = list(map(lambda wc: wc.cluster, result))
        return result

    class WrappedCluster(object):
        """
        Wrapper for clusters which stores additional data required for clustering (internal use only)
        """
        def __init__(self, cluster, idx, clusterer: "GreedyAgglomerativeClustering"):
            self.merged_into_cluster: Optional[GreedyAgglomerativeClustering.WrappedCluster] = None
            self.merges = []
            self.cluster = cluster
            self.idx = idx
            self.clusterer = clusterer

        def is_merged(self) -> bool:
            return self.merged_into_cluster is not None

        def get_cluster_association(self) -> "GreedyAgglomerativeClustering.WrappedCluster":
            """
            Gets the wrapped cluster that this cluster's points have ultimately been merged into (which may be the cluster itself)

            :return: the wrapped cluster this cluster's points are associated with
            """
            if self.merged_into_cluster is None:
                return self
            else:
                return self.merged_into_cluster.get_cluster_association()
            
        def remove_merges(self):
            for merge in self.merges:
                merge.evaporated = True
            self.merges = []

        def compute_merges(self, initial: bool, merged_cluster_indices: Tuple[int, int] = None):
            # add new merges to queue
            wrapped_clusters = self.clusterer.wrapped_clusters
            for item in self.clusterer.mergeCandidateDeterminationStrategy.iter_candidate_indices(self, initial, merged_cluster_indices):
                merge: Optional[GreedyAgglomerativeClustering.ClusterMerge] = None
                if type(item) == int:
                    other_idx = item
                    if other_idx != self.idx:
                        other = wrapped_clusters[other_idx]
                        if not other.is_merged():
                            merge_cost = self.cluster.merge_cost(other.cluster)
                            if not math.isinf(merge_cost):
                                merge = GreedyAgglomerativeClustering.ClusterMerge(self, other, merge_cost)
                else:
                    merge = item
                    assert merge.c1.idx == self.idx
                if merge is not None:
                    merge.c1.merges.append(merge)
                    merge.c2.merges.append(merge)
                    self.clusterer.prioritised_merges.put(merge)

        def __str__(self):
            return "Cluster[idx=%d]" % self.idx
        
    class ClusterMerge(object):
        """
        Represents a potential merge
        """
        log = log.getChild(__qualname__)

        def __init__(self, c1: "GreedyAgglomerativeClustering.WrappedCluster", c2: "GreedyAgglomerativeClustering.WrappedCluster",
                merge_cost):
            self.c1 = c1
            self.c2 = c2
            self.merge_cost = merge_cost
            self.evaporated = False

        def apply(self):
            c1, c2 = self.c1, self.c2
            self.log.debug("Merging %s into %s..." % (str(c1), str(c2)))
            c1.cluster.merge(c2.cluster)
            c2.merged_into_cluster = c1
            c1.remove_merges()
            c2.remove_merges()
            self.log.debug("Computing new merge costs for %s..." % str(c1))
            c1.compute_merges(False, merged_cluster_indices=(c1.idx, c2.idx))
        
        def __lt__(self, other):
            return self.merge_cost < other.merge_cost

    class MergeCandidateDeterminationStrategy(ABC):
        def __init__(self):
            self.clusterer: Optional["GreedyAgglomerativeClustering"] = None

        """
        Determines the indices of clusters which should be evaluated with regard to their merge costs
        """
        def set_clusterer(self, clusterer: "GreedyAgglomerativeClustering"):
            """
            Initialises the clusterer the strategy is applied to
            :param clusterer: the clusterer
            """
            self.clusterer = clusterer

        @abstractmethod
        def iter_candidate_indices(self, wc: "GreedyAgglomerativeClustering.WrappedCluster", initial: bool,
                merged_cluster_indices: Tuple[int, int] = None) -> Iterator[Union[int, "GreedyAgglomerativeClustering.ClusterMerge"]]:
            """
            :param wc: the wrapped cluster: the cluster for which to determine the cluster indices that are to be considered for
                a potential merge
            :param initial: whether we are computing the initial candidates (at the start of the clustering algorithm)
            :param merged_cluster_indices: [for initial=False] the pair of cluster indices that were just joined to form the updated
                cluster wc
            :return: an iterator of cluster indices that should be evaluated as potential merge partners for wc (it may contain the
                index of wc, which will be ignored)
            """
            pass

    class MergeCandidateDeterminationStrategyDefault(MergeCandidateDeterminationStrategy):
        def iter_candidate_indices(self, wc: "GreedyAgglomerativeClustering.WrappedCluster", initial: bool,
                merged_cluster_indices: Tuple[int, int] = None) -> Iterator[Union[int, "GreedyAgglomerativeClustering.ClusterMerge"]]:
            n = len(self.clusterer.wrapped_clusters)
            if initial:
                return range(wc.idx + 1, n)
            else:
                return range(n)



from typing import Union, Callable

import torch
from torch import nn
from torch.nn import functional as F

from ...torch_base import MCDropoutCapableNNModule
from ...torch_enums import ActivationFunction


class LSTNetwork(MCDropoutCapableNNModule):
    """
    Network for (auto-regressive) time-series prediction with long- and short-term dependencies as proposed by G. Lai et al.
    It applies two parallel paths to a time series of size (numInputTimeSlices, inputDimPerTimeSlice):
    
        * Complex path with the following stages:
        
            * Convolutions on the time series input data (CNNs):
              For a CNN with numCnnTimeSlices (= kernel size), it produces an output series of size numInputTimeSlices-numCnnTimeSlices+1.
              If the number of parallel convolutions is numConvolutions, the total output size of this stage is thus
              numConvolutions*(numInputTimeSlices-numCnnTimeSlices+1)
            * Two RNN components which process the CNN output in parallel:
            
                * RNN (GRU)
                  The output dimension of this stage is the hidden state of the GRU after seeing the entire
                  input data from the previous stage, i.e. if has size hidRNN.
                * Skip-RNN (GRU), which processes time series elements that are 'skip' time slices apart.
                  It does this by grouping the input such that 'skip' GRUs are applied in parallel, which all use the same parameters.
                  If the hidden state dimension of each GRU is hidSkip, then the output size of this stage is skip*hidSkip.
                  
            * Dense layer
            
        * Direct regression dense layer (so-called "highway" path).
        
    The model ultimately combines the outputs of these two paths via a combination function.
    Many parts of the model are optional and can be completely disabled.
    The model can produce one or more (potentially multi-dimensional) outputs, where each output typically corresponds
    to a time slice for which a prediction is made.

    The model expects as input a tensor of size (batchSize, numInputTimeSlices, inputDimPerTimeSlice).
    As output, the model will produce a tensor of size (batchSize, numOutputTimeSlices, outputDimPerTimeSlice)
    if isClassification==False (default) and a tensor of size (batchSize, outputDimPerTimeSlice=numClasses, numOutputTimeSlices)
    if isClassification==True; the latter shape matches what is required by the multi-dimensional case of loss function
    CrossEntropyLoss, for example, and therefore is suitable for classification use cases.
    """
    def __init__(self, numInputTimeSlices, inputDimPerTimeSlice, numOutputTimeSlices=1, outputDimPerTimeSlice=1,
            numConvolutions: int = 100, numCnnTimeSlices: int = 6, hidRNN: int = 100, skip: int = 0, hidSkip: int = 5,
            hwWindow: int = 0, hwCombine: str = "plus", dropout=0.2, outputActivation: Union[str, ActivationFunction, Callable] = "sigmoid",
            isClassification=False):
        """
        :param numInputTimeSlices: the number of input time slices
        :param inputDimPerTimeSlice: the dimension of the input data per time slice
        :param numOutputTimeSlices: the number of time slices predicted by the model
        :param outputDimPerTimeSlice: the number of dimensions per output time slice. While this is the number of
            target variables per time slice for regression problems, this must be the number of classes for classification problems.
        :param numCnnTimeSlices: the number of time slices considered by each convolution (i.e. it is one of the dimensions of the matrix used for
            convolutions, the other dimension being inputDimPerTimeSlice), a.k.a. "Ck"
        :param numConvolutions: the number of separate convolutions to apply, i.e. the number of independent convolution matrices, a.k.a "hidC";
            if it is 0, then the entire complex processing path is not applied.
        :param hidRNN: the number of hidden output dimensions for the RNN stage
        :param skip: the number of time slices to skip for the skip-RNN. If it is 0, then the skip-RNN is not used.
        :param hidSkip: the number of output dimensions of each of the skip parallel RNNs
        :param hwWindow: the number of time slices from the end of the input time series to consider as input for the highway component.
            If it is 0, the highway component is not used.
        :param hwCombine: {"plus", "product", "bilinear"} the function with which the highway component's output is combined with the complex path's output
        :param dropout: the dropout probability to use during training (dropouts are applied after every major step in the evaluation path)
        :param outputActivation: the output activation function
        :param isClassification: whether the model is to serve as a classifier, in which case the output tensor dimension ordering is adapted
            to suit loss functions such as CrossEntropyLoss
        """
        if numConvolutions == 0 and hwWindow == 0:
            raise ValueError("No processing paths remain")
        if numInputTimeSlices < numCnnTimeSlices or (hwWindow != 0 and hwWindow < numInputTimeSlices):
            raise Exception("Inconsistent numbers of times slices provided")

        super().__init__()
        self.inputDimPerTimeSlice = inputDimPerTimeSlice
        self.timeSeriesDimPerTimeSlice = outputDimPerTimeSlice
        self.totalOutputDim = self.timeSeriesDimPerTimeSlice * numOutputTimeSlices
        self.numOutputTimeSlices = numOutputTimeSlices
        self.window = numInputTimeSlices
        self.hidRNN = hidRNN
        self.numConv = numConvolutions
        self.hidSkip = hidSkip
        self.Ck = numCnnTimeSlices  # the "height" of the CNN filter/kernel; the "width" being inputDimPerTimeSlice
        self.convSeqLength = self.window - self.Ck + 1  # the length of the output sequence produced by the CNN for each kernel matrix
        self.skip = skip
        self.hw = hwWindow
        self.pDropout = dropout
        self.isClassification = isClassification

        # configure CNN-RNN path
        if self.numConv > 0:
            self.conv1 = nn.Conv2d(1, self.numConv, kernel_size=(self.Ck, self.inputDimPerTimeSlice))  # produce numConv sequences using numConv kernel matrices of size (height=Ck, width=inputDimPerTimeSlice)
            self.GRU1 = nn.GRU(self.numConv, self.hidRNN)
            if self.skip > 0:
                self.skipRnnSeqLength = self.convSeqLength // self.skip  # we divide by skip to obtain the sequence length, because, in order to support skipping via a regrouping of the tensor, the Skip-RNN processes skip entries of the series in parallel to produce skip hidden output vectors
                if self.skipRnnSeqLength == 0:
                    raise Exception("Window size %d is not large enough for skip length %d; would result in Skip-RNN sequence length of 0!" % (self.window, self.skip))
                self.GRUskip = nn.GRU(self.numConv, self.hidSkip)
                self.linear1 = nn.Linear(self.hidRNN + self.skip * self.hidSkip, self.totalOutputDim)
            else:
                self.linear1 = nn.Linear(self.hidRNN, self.totalOutputDim)

        # configure highway component
        if self.hw > 0:
            # direct mapping from all inputs to all outputs
            self.highway = nn.Linear(self.hw * self.inputDimPerTimeSlice, self.totalOutputDim)
            if hwCombine == 'plus':
                self.highwayCombine = self._plus
            elif hwCombine == 'product':
                self.highwayCombine = self._product
            elif hwCombine == 'bilinear':
                self.highwayCombine = nn.Bilinear(self.totalOutputDim, self.totalOutputDim, self.totalOutputDim)
            else:
                raise ValueError("Unknown highway combination function '%s'" % hwCombine)

        self.output = ActivationFunction.torch_function_from_any(outputActivation)

    def forward(self, x):
        batch_size = x.size(0)
        # x has size (batch_size, window=numInputTimeSlices, inputDimPerTimeSlice)

        dropout = lambda x: self._dropout(x, p_training=self.pDropout, p_inference=self.pDropout)

        res = None

        if self.numConv > 0:
            # CNN
            # convSeqLength = self.window - self.Ck + 1
            # convolution produces, via numConv kernel matrices of dimension (height=Ck, width=inputDimPerTimeSlice), from an original input sequence of length window, numConv output sequences of length convSeqLength
            c = x.view(batch_size, 1, self.window, self.inputDimPerTimeSlice)  # insert one dim of size 1 (one channel): (batch_size, 1, height=window, width=inputDimPerTimeSlice)
            c = F.relu(self.conv1(c))  # (batch_size, channels=numConv, convSeqLength, 1)
            c = dropout(c)
            c = torch.squeeze(c, 3)  # drops last dimension, i.e. new size (batch_size, numConv, convSeqLength)

            # RNN
            # It processes the numConv sequences of length convSeqLength obtained through convolution and keep the hidden state at the end, which is comprised of hidR entries
            # Specifically, it squashes the numConv sequences of length convSeqLength to a vector of size hidS (by iterating through the sequences and applying the same model in each step, processing all batches in parallel)
            r = c.permute(2, 0, 1).contiguous()  # (convSeqLength, batch_size, numConv)
            self.GRU1.flatten_parameters()
            _, r = self.GRU1(r)  # maps (seq_len=convSeqLength, batch=batch_size, input_size=numConv) -> hidden state (num_layers=1, batch=batch_size, hidden_size=hidR)
            r = torch.squeeze(r, 0)  # (batch_size, hidR)
            r = dropout(r)

            # Skip-RNN
            if self.skip > 0:
                s = c[:, :, -(self.skipRnnSeqLength * self.skip):].contiguous()  # (batch_size, numConv, convSeqLength) -> (batch_size, numConv, skipRnnSeqLength * skip)
                s = s.view(batch_size, self.numConv, self.skipRnnSeqLength, self.skip)  # (batch_size, numConv, skipRnnSeqLength, skip)
                s = s.permute(2, 0, 3, 1).contiguous()  # (skipRnnSeqLength, batch_size, skip, numConv)
                s = s.view(self.skipRnnSeqLength, batch_size * self.skip, self.numConv)  # (skipRnnSeqLength, batch_size * skip, numConv)
                # Why the above view makes sense:
                # skipRnnSeqLength is the sequence length considered for the RNN, i.e. the number of steps that is taken for each sequence.
                # The batch_size*skip elements of the second dimension are all processed in parallel, i.e. there are batch_size*skip RNNs being applied in parallel.
                # By scaling the actual batch size with 'skip', we process 'skip' RNNs of each batch in parallel, such that each RNN consecutively processes entries that are 'skip' steps apart
                self.GRUskip.flatten_parameters()
                _, s = self.GRUskip(s)  # maps (seq_len=skipRnnSeqLength, batch=batch_size * skip, input_size=numConv) -> hidden state (num_layers=1, batch=batch_size * skip, hidden_size=hidS)
                # Because of the way the data is grouped, we obtain not one vector of size hidS but skip vectors of size hidS
                s = s.view(batch_size, self.skip * self.hidSkip)  # regroup by batch -> (batch_size, skip * hidS)
                s = dropout(s)
                r = torch.cat((r, s), 1)  # (batch_size, hidR + skip * hidS)

            res = self.linear1(r)  # (batch_size, totalOutputDim)

        # auto-regressive highway model
        if self.hw > 0:
            resHW = x[:, -self.hw:, :]  # keep only the last hw entries for each input: (batch_size, hw, inputDimPerTimeSlice)
            resHW = resHW.view(-1, self.hw * self.inputDimPerTimeSlice)  # (batch_size, hw * inputDimPerTimeSlice)
            resHW = self.highway(resHW)  # (batch_size, totalOutputDim)
            if res is None:
                res = resHW
            else:
                res = self.highwayCombine(res, resHW)  # (batch_size, totalOutputDim)

        if self.output:
            res = self.output(res)

        res = res.view(batch_size, self.numOutputTimeSlices, self.timeSeriesDimPerTimeSlice)
        if self.isClassification:
            res = res.permute(0, 2, 1)
        return res

    @staticmethod
    def _plus(x, y):
        return x + y

    @staticmethod
    def _product(x, y):
        return x * y
